import datetime
import os
from typing import TYPE_CHECKING
from typing import TypedDict

from django.core.exceptions import SuspiciousFileOperation
from django.core.files.storage import Storage
from django.urls import path
from django.utils.crypto import get_random_string

if TYPE_CHECKING:
    from alliance_platform.storage.async_uploads.registry import AsyncFieldRegistry


class GenerateUploadUrlResponse(TypedDict):
    #: The URL to post to
    url: str
    #: A dictionary of form field names and their values to be included when submitting
    fields: dict


class AsyncUploadStorage(Storage):
    """Base storage class for use with :class:`~alliance_platform.storage.async_uploads.models.AsyncFileField`

    Provides ``generate_upload_url`` to return a URL to upload a file to (eg. a signed URL) and a ``move_file`` method
    to move a file from a temporary location to the permanent location. The key used for temporary files is returned
    by ``generate_temporary_path`` and ``is_temporary_path`` should return whether a file is in the temporary location
    and needs to be moved to a permanent location.

    See :class:`~alliance_platform.storage.async_uploads.models.AsyncFileMixin` for a detailed explanation of how
    all the pieces fit together.
    """

    #: The prefix to use on files when first uploaded before they are moved. This is used to identify a file as a
    #: temporary file to know whether to move them and so *must* be unique to temporary files.
    temporary_key_prefix = "async-temp-files"

    def generate_upload_url(self, name: str, field_id: str, *args, **kwargs) -> GenerateUploadUrlResponse:
        """Should return a URL that a file can be uploaded directly to

        In S3 this would be a signed URL.
        """
        raise NotImplementedError("generate_upload_url must be implemented")

    def generate_download_url(self, key: str, field_id: str, **kwargs):
        """Should return a URL that the specified key should be downloadable from

        In S3 this would be a signed URL.
        """
        raise NotImplementedError("generate_download_url must be implemented")

    def move_file(self, from_key, to_key):
        """Move a file from one location to another

        The details of this depend on the storage solution. In S3 this involves copying the file
        and then deleting the original file.
        """
        raise NotImplementedError("move_file must be implemented")

    def generate_temporary_path(self, filename, max_length=None):
        """Generates a unique key to upload ``filename`` to

        This generates a string like ``async-temp-files/2021/03/03/fVy5cSVBQpOb-test.png`` where
        ``test.png`` is the ``filename`` passed in. The part after ``-`` will be truncated to fit within
        ``max_length`` but will retain the file extension. If there is insufficient length to
        accommodate the temporary path prefix (up to the ``-`` and the file extension an error
        will be thrown.
        """
        today = datetime.date.today().strftime("%Y/%m/%d")
        secret = get_random_string(12)
        prefix = f"{self.temporary_key_prefix}/{today}/{secret}-"
        if max_length:
            file_root, file_ext = os.path.splitext(filename)
            length_remaining = max_length - len(prefix) - len(file_ext)
            if length_remaining < 0:
                raise SuspiciousFileOperation(
                    f'Storage can not find an available temp filename for {filename}". '
                    "Please make sure that the corresponding file field "
                    'allows sufficient "max_length".'
                )
            truncation = len(file_root) - length_remaining
            if truncation > 0:
                file_root = file_root[:-truncation]
            filename = f"{file_root}{file_ext}"
        return f"{prefix}{filename}"

    def is_temporary_path(self, filename: str) -> bool:
        """Is the specified path that of a temporary file?

        This is used to determine whether a file should be moved to a permanent location. If this
        returns true it's expected an AsyncTempFile exists with the matching key.

        Default implementation checks if path begins with :code:`temporary_key_prefix`
        """
        if not filename:
            return False
        return filename.startswith(self.temporary_key_prefix + "/")

    def get_url_patterns(self, registry: "AsyncFieldRegistry"):
        """Return the URL patterns for any views required by the storage class

        When extending ``AsyncUploadStorage``, this method can be implemented if any custom views
        are required to support the implementation. By default, two views are supplied:

        1) :class:`~alliance_platform.storage.async_uploads.views.DownloadRedirectView` to support downloading an existing file. This
           is attached to the `"download-file/"` path.
        2) :class:`~alliance_platform.storage.async_uploads.views.GenerateUploadUrlView` to generate a URL that can be uploaded to
           directly from the frontend. This is attached to the `"generate-upload-url/"` path.

        This method is called by :meth:`~alliance_platform.storage.async_uploads.registry.AsyncFieldRegistry.get_url_patterns`.
        """
        from alliance_platform.storage.async_uploads.views import DownloadRedirectView
        from alliance_platform.storage.async_uploads.views import GenerateUploadUrlView

        # already generated patterns for these views, can return nothing
        if registry.attached_download_view and registry.attached_view:
            return []

        return [
            path("download-file/", DownloadRedirectView.as_view(registry=registry)),
            path("generate-upload-url/", GenerateUploadUrlView.as_view(registry=registry)),
        ]



import os
from typing import TYPE_CHECKING
from urllib.parse import urlencode

from alliance_platform.storage.async_uploads.storage.base import AsyncUploadStorage
from alliance_platform.storage.async_uploads.storage.base import GenerateUploadUrlResponse
from django.core.files.move import file_move_safe
from django.core.files.storage import FileSystemStorage
from django.core.signing import TimestampSigner
from django.urls import path
from django.urls import reverse

if TYPE_CHECKING:
    # circular imports
    from alliance_platform.storage.async_uploads.registry import AsyncFieldRegistry


class FileSystemAsyncUploadStorage(FileSystemStorage, AsyncUploadStorage):
    """Implementation of AsyncUploadStorage that uploads directly to the local server

    This is useful in local dev, or when you still want the behaviour of uploading immediately rather than waiting
    until the whole form is submitted.

    To use this by default, set the :setting:`STORAGES <django:STORAGES>` setting::

        STORAGES = {
            "default": {
                "BACKEND": "alliance_platform.storage.async_uploads.storage.filesystem.FileSystemAsyncUploadStorage"
            },
        }

    Alternatively, pass an instance of the class to the ``storage`` argument on :class:`~alliance_platform.storage.async_uploads.models.AsyncFileField`
    or :class:`~alliance_platform.storage.async_uploads.models.AsyncImageField`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.signer = TimestampSigner()
        # This is only available in >= Django 5.1.
        if not hasattr(self, "_allow_overwrite"):
            self._allow_overwrite = False

    def generate_upload_url(self, name: str, field_id: str, *args, **kwargs) -> GenerateUploadUrlResponse:
        """
        Generates a signed URL that the frontend can use to upload a file directly to the server.
        """
        signed_path = self.signer.sign(name)
        query_params = urlencode({"path": signed_path, "field_id": field_id})
        upload_url = reverse("file_system_async_storage_upload") + f"?{query_params}"
        return {"url": upload_url, "fields": {}}

    def generate_download_url(self, key: str, field_id: str, **kwargs):
        signed_path = self.signer.sign(key)
        query_params = urlencode({"path": signed_path, "field_id": field_id})
        upload_url = reverse("file_system_async_storage_download") + f"?{query_params}"
        return upload_url

    def move_file(self, from_key, to_key):
        """
        Moves a file from the temporary location to the permanent location.
        """
        # Create any intermediate directories that do not exist.
        full_path = self.path(to_key)

        # This logic for creating dirs is taken from FileSystemStorage._save
        directory = os.path.dirname(full_path)
        try:
            if self.directory_permissions_mode is not None:
                # Set the umask because os.makedirs() doesn't apply the "mode"
                # argument to intermediate-level directories.
                old_umask = os.umask(0o777 & ~self.directory_permissions_mode)
                try:
                    os.makedirs(directory, self.directory_permissions_mode, exist_ok=True)
                finally:
                    os.umask(old_umask)
            else:
                os.makedirs(directory, exist_ok=True)
        except FileExistsError:
            raise FileExistsError("%s exists and is not a directory." % directory)
        file_move_safe(self.path(from_key), full_path, allow_overwrite=self._allow_overwrite)

    def get_url_patterns(self, registry: "AsyncFieldRegistry"):
        patterns = super().get_url_patterns(registry)
        from alliance_platform.storage.async_uploads.views.filesystem import (
            FileSystemAsyncStorageDownloadView,
        )
        from alliance_platform.storage.async_uploads.views.filesystem import FileSystemAsyncStorageUploadView

        # already generated patterns for these views, can return nothing
        if getattr(registry, FileSystemAsyncStorageUploadView.registration_attach_key, None):
            return patterns

        return [
            *patterns,
            path(
                "upload-file-direct/",
                FileSystemAsyncStorageUploadView.as_view(registry=registry),
                name="file_system_async_storage_upload",
            ),
            path(
                "download-file-direct/",
                FileSystemAsyncStorageDownloadView.as_view(registry=registry),
                name="file_system_async_storage_download",
            ),
        ]



from datetime import timedelta
from typing import cast

from alliance_platform.storage.async_uploads.models import AsyncFileField
from alliance_platform.storage.async_uploads.models import AsyncTempFile
from django.core.management import BaseCommand
from django.db import models
from django.utils import timezone


class Command(BaseCommand):
    """Command to clean old AsyncTempFile records that accumulate if a file is uploaded but form not submitted

    Usage::

        ./manage.py cleanup_async_temp_files

        # Specify age in hours - anything older is deleted
        ./manage.py cleanup_async_temp_files --age=24
    """

    help = "Cleanup old AsyncTempFile records that accumulate if a file is uploaded but form not submitted"

    def add_arguments(self, parser):
        parser.add_argument(
            "--age",
            action="store",
            default=48,
            type=int,
            help="Only files older than this will be removed (in hours). Defaults to 48.",
        )

        parser.add_argument(
            "--quiet",
            action="store_true",
            default=False,
            help="Don't output number of items removed to stdout",
        )

    def handle(self, *app_labels, **options):
        temp_files = AsyncTempFile.objects.filter(
            created_at__lt=timezone.now() - timedelta(hours=options["age"]), error__isnull=True
        )
        completed_files = temp_files.exclude(moved_to_location="")
        if not options["quiet"]:
            self.stdout.write(
                self.style.NOTICE(f"Found {completed_files.count()} completed temp files for removal")
            )
        completed_files.delete()
        if not options["quiet"]:
            self.stdout.write(
                self.style.NOTICE(f"Found {temp_files.count()} incomplete temp files for removal")
            )
        for temp_file in temp_files:
            field = cast(models.Model, temp_file.content_type.model_class())._meta.get_field(
                temp_file.field_name
            )
            try:
                cast(AsyncFileField, field).storage.delete(temp_file.key)
            except Exception as e:
                if not options["quiet"]:
                    # Ignore any errors from storage and continue processing
                    self.stdout.write(
                        self.style.ERROR(f"Storage failed to remove key {temp_file.key}: {str(e)}")
                    )
            temp_file.delete()



import fnmatch

from django.contrib.staticfiles.storage import ManifestStaticFilesStorage
from django.core.exceptions import ImproperlyConfigured


class ExcludingManifestStaticFilesStorage(ManifestStaticFilesStorage):
    """
    Custom static files storage that excludes specified files from being processed by
    :class:`~django:django.contrib.staticfiles.storage.ManifestStaticFilesStorage` (i.e., they won't have their names hashed).
    This is useful for files that are already hashed, or are never cached by the browser.

    To use this class, set it in the :setting:`STORAGES <django:STORAGES>` setting. Pass
    the ``exclude_patterns`` option as shown below. This can be set to patterns accepted
    by :external:py:func:`~fnmatch.fnmatch`

    .. note::

        The patterns accepted are not the same as regular expressions - see :external:py:mod:`fnmatch` for details.

    Usage::

        STORAGES = {
            "staticfiles": {
                "BACKEND": "alliance_platform.storage.staticfiles.storage.ExcludingManifestStaticFilesStorage",
                "OPTIONS": {"exclude_patterns": ["frontend/build/*", "*.pdf"]},
            }
        }
    """

    def __init__(self, *args, exclude_patterns=None, **kwargs):
        if exclude_patterns is None:
            raise ImproperlyConfigured(
                "ExcludingManifestStaticFilesStorage requires 'exclude_patterns' option to be set. This can be passed under the STORAGES['staticfiles']['OPTIONS']['exclude_patterns'] setting."
            )
        super().__init__(*args, **kwargs)
        self._exclude_patterns = exclude_patterns or []

    def should_exclude(self, name):
        """
        Check if the file name matches any of the exclude patterns.
        """
        for pattern in self._exclude_patterns:
            if fnmatch.fnmatch(name, pattern):
                return True
        return False

    def hashed_name(self, name, content=None, filename=None):
        """
        Override the hashed_name method to skip hashing for excluded files.
        """
        if self.should_exclude(name):
            # Return the original name without hashing
            return name
        return super().hashed_name(name, content, filename)

    def post_process(self, paths, **options):
        """
        Don't post process excluded files at all. Note that we still need ``hashed_name``
        above for any excluded files that are referenced in non-excluded files.
        """
        paths_to_hash = {}
        other_paths = []
        for path in paths:
            if self.should_exclude(path):
                other_paths.append(path)
            else:
                paths_to_hash[path] = paths[path]
        yield from super().post_process(paths_to_hash, **options)
        # add an identity mapping for excluded items otherwise base class will throw
        # an error when ``manifest_strict`` is ``True``.
        for path in other_paths:
            self.hashed_files[path] = path



# Generated by Django 5.1 on 2024-09-25 04:04

from django.db import migrations
from django.db import models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="AsyncTempFile",
            fields=[
                (
                    "id",
                    models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("original_filename", models.TextField()),
                ("key", models.CharField(max_length=500)),
                ("field_name", models.CharField(max_length=100)),
                ("moved_to_location", models.TextField(blank=True)),
                ("error", models.TextField(null=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="contenttypes.contenttype"
                    ),
                ),
            ],
            options={
                "db_table": "alliance_platform_storage_async_temp_file",
            },
        ),
    ]



from typing import cast

from alliance_platform.storage.async_uploads.forms import AsyncFileInputDataLengthValidator
from alliance_platform.storage.async_uploads.forms import AsyncFileInputDataValidator
from alliance_platform.storage.async_uploads.models import AsyncFieldFile
from alliance_platform.storage.async_uploads.models import AsyncFileField as AsyncFileBaseField
from alliance_platform.storage.async_uploads.models import AsyncFileInputData
from django.core.exceptions import ImproperlyConfigured
from django.core.exceptions import ValidationError as CoreValidationError

try:
    from rest_framework import serializers
    from rest_framework.fields import empty
    from rest_framework.fields import get_attribute
    from rest_framework.utils.formatting import lazy_format
except ImportError as e:
    raise ImproperlyConfigured(
        "Optional dependency 'rest_framework' is not installed. This is required to make use of the DRF fields."
    ) from e


class AsyncFileField(serializers.ModelField):
    """Field that works with :class:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage` to handle uploading directly to
    external service like S3 or Azure.

    Unlike most fields this field must be backed by an underlying model field specified in the ``model_field``
    kwarg. This is inferred automatically if using a base ModelSerializer class with the following:

    .. code:: python

        import alliance_platform.storage.async_uploads.models as async_file_fields
        class BaseModelSerializer(ModelSerializer):
            serializer_field_mapping = {
                **ModelSerializer.serializer_field_mapping,
                async_file_fields.AsyncFileField: AsyncFileField,
                async_file_fields.AsyncImageField: AsyncImageField,
            }

    This field expects to receive data in the shape { "key": "/storage/file.png", "name": "file.png" }. It will
    extract the key to be set on the file field. See :class:`~alliance_platform.storage.async_uploads.models.AsyncFileMixin`
    for more details. For :class:`~alliance_platform.storage.async_uploads.rest_framework.AsyncImageField` the ``width`` and ``height``
    keys may also exist in the data. See :class:`~alliance_platform.async_uploads.models.AsyncFileInputData`

    Serialized data is in shape  { "key": "/storage/file.png", "name": "file.png", "url": "/download/?field_id=..." }
    which is supported by the UploadWidget on the frontend.
    """

    default_error_messages = {
        "max_length": "Ensure this filename has at most {max_length} characters.",
        "invalid_key": "{error}",
    }

    def __init__(self, *args, **kwargs):
        max_length = kwargs.pop("max_length", None)
        # ModelField handles max_length if provided - bypass this so we can use custom validator
        super().__init__(*args, **kwargs)
        self.max_length = max_length
        if self.max_length is not None:
            format_str = cast(str, self.error_messages["max_length"])
            message = lazy_format(format_str, max_length=self.max_length)
            self.validators.append(
                AsyncFileInputDataLengthValidator(int(self.max_length), message=cast(str, message))
            )
        self.validators.append(AsyncFileInputDataValidator())

    def get_value(self, dictionary):
        """Convert incoming value into AsyncFileInputData"""
        value = super().get_value(dictionary)
        if not value or value == empty:
            return value
        if not isinstance(value, dict):
            return value

        parent_meta = self.parent.Meta
        pk_name = parent_meta.model._meta.pk.name
        if pk_name in dictionary:
            self._instance_pk = dictionary[pk_name]

        return AsyncFileInputData.create_from_user_input(value)

    def to_internal_value(self, value: AsyncFileInputData):
        if isinstance(value, AsyncFileInputData):
            value.update_dimension_cache(self.model_field)
            existing_value = (
                None if not self.parent.instance else getattr(self.parent.instance, self.model_field.name)
            )

            if not existing_value and (nested_pk := getattr(self, "_instance_pk", None)):
                parent_meta = self.parent.Meta  # type: ignore[attr-defined] # metaclass
                instance = parent_meta.model._base_manager.get(pk=nested_pk)
                existing_value = getattr(instance, self.model_field.name)

            try:
                value.validate_key(
                    cast(AsyncFileBaseField, self.model_field),
                    cast(str | AsyncFieldFile | None, existing_value),
                )
            except CoreValidationError as e:
                self.fail("invalid_key", error=e)
        return value

    def to_representation(self, value):
        if self.source_attrs:
            # if source is set to e.g. fk1.fk2.file_field, we need to that final relation 'fk2'
            # in the chain (but not the field - that's handled by value_from_object below)
            value = get_attribute(value, self.source_attrs[:-1])
        value = self.model_field.value_from_object(value)
        if not value:
            return None
        try:
            url = value.url
        except Exception:
            # url isn't critical; if it fails ignore it and frontend may just not be able
            # to show a thumbnail
            url = None
        return {"key": value.name, "name": value.name.split("/")[-1], "url": url}


class AsyncImageField(AsyncFileField):
    """Same behaviour as :class:`~alliance_platform.storage.async_uploads.rest_framework.AsyncFileField`

    This exists as a separate class to allow codegen to differentiate for the purpose
    of customising the field & widget on the frontend."""

    pass



from __future__ import annotations

from dataclasses import dataclass
from hashlib import shake_256
import logging
import mimetypes
from typing import TYPE_CHECKING
from typing import Protocol
from typing import TypedDict
from typing import cast
from urllib.parse import urlencode

from alliance_platform.core.auth import resolve_perm_name
from alliance_platform.storage.async_uploads.registry import default_async_field_registry
from alliance_platform.storage.async_uploads.storage.base import AsyncUploadStorage
from allianceutils.middleware import CurrentRequestMiddleware
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db import transaction
from django.db.models import FileField
from django.db.models import ImageField
from django.db.models import Model
from django.db.models import Q
from django.db.models import signals
from django.db.models.fields.files import FieldFile
from django.db.models.fields.files import FileDescriptor
from django.db.models.fields.files import ImageFieldFile
from django.urls import reverse

logger = logging.getLogger("alliance_platform.storage")


if TYPE_CHECKING:
    # this is to avoid accidentally importing the incorrect classes that are
    # only in place for type checking

    class AsyncFileMixinProtocol(Protocol):
        model: type[models.Model]

        def contribute_to_class(self, cls: type[Model], name: str, private_only: bool = ...) -> None: ...

        def __init__(self, *args, max_length: int | None, **kwargs) -> None: ...

    class FieldFileMixinProtocol(Protocol):
        instance: models.Model
        name: str | None
        field: AsyncFileField | AsyncImageField

else:

    class AsyncFileMixinProtocol:
        pass

    class FieldFileMixinProtocol:
        pass


class AsyncFileModelRegistry:
    """Class that tracks files on a model and handles moving files on save

    When the save occurs all temporary files will be moved to final location and a
    single save will be done on the record to write the final field values to the db.
    All AsyncTempFiles that are no longer needed will be deleted in a single query.

    If any error occurs on a move then it is logged to the alliance_platform.storage logger and
    the AsyncTempFile is retained with the ``error`` field set. This can then be manually
    followed up. The cleanup_async_temp_files management command will not remove these
    files.
    """

    _save_in_progress: list[Model]

    def __init__(self):
        self._save_in_progress = []

    def _move_temp_files_into_place(self, instance: Model, *args, **kwargs):
        """For each async file on the instance move them from temporary location to permanent location

        This is triggered by the ``post_save`` signal. The destination filename goes through the normal
        FileField logic using ``AsyncTempFile.original_filename``. If ``FileField`` has ``upload_to`` set
        then this will be used, otherwise the destination will be the same as ``original_filename``.

        If the destination filename is too long for the database after ``upload_to`` is applied then the
        file will be truncated by the ``Storage.get_available_name`` method.

        Also see :meth:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.move_file`

        This uses as few queries as possible to do the moves and update new values. Only async fields
        that have a temporary key are touched - other fields that have already been moved are left
        alone.
        """
        model_class = instance.__class__
        class_fields = model_class._meta.fields

        fields = [f for f in class_fields if isinstance(f, AsyncFileMixin)]

        if not fields:
            return

        # Use ``is`` like this instead of ``instance in self._save_in_progress`` so we check object is same not just _pk
        if [x for x in self._save_in_progress if x is instance]:
            return
        _filter = Q()
        for field in fields:
            file = getattr(instance, field.name)
            _filter |= Q(
                key=file.name,
                content_type=ContentType.objects.get_for_model(field.model),
                field_name=field.name,
            )
        with transaction.atomic():
            temp_files_qs = AsyncTempFile.objects.filter(_filter).select_for_update()
            temp_files = {file.field_name: file for file in temp_files_qs}
            pending_save: list[AsyncTempFile] = []
            should_save = False
            for field in fields:
                file = getattr(instance, field.name)
                field.before_process_file(instance)
                if field.storage.is_temporary_path(file.name):
                    try:
                        temp_file = temp_files[field.name]
                    except KeyError:
                        # This is unexpected. Could mean file was cleaned up manually before submission occurred.
                        logger.error(
                            f"{file.name} is a temporary file path in {field.storage.__class__.__name__} but no matching "
                            f"AsyncTempFile with that key could be found. File cannot be moved to its permanent location. "
                            f"See record {instance.__class__.__name__}(pk={instance.pk})"
                        )
                    else:
                        try:
                            if temp_file.moved_to_location:
                                file.name = temp_file.moved_to_location
                                try:
                                    request = CurrentRequestMiddleware.get_request()
                                    url = request.get_full_path() if request else "(unknown)"
                                except Exception:
                                    url = "(unknown)"
                                logger.warning(
                                    f"Temp file for has already been moved to its final location using stored destination for field value ('{file.name}'). This happens "
                                    "when a form is submitted multiple times and likely indicates a bug in the form that should be resolved. "
                                    f"Error occurred on URL {url}. See record {instance.__class__.__name__}(pk={instance.pk}), field '{field.name}'"
                                )
                            else:
                                target_path = cast(models.FileField, field).generate_filename(
                                    instance, temp_file.original_filename
                                )
                                if field.max_length:
                                    target_path = field.storage.get_available_name(
                                        target_path, max_length=field.max_length
                                    )
                                field.storage.move_file(file.name, target_path)
                                file.name = target_path
                                temp_file.moved_to_location = target_path
                                pending_save.append(temp_file)
                        except Exception:
                            logger.error(
                                f"{file.name} is a temporary file path in {field.storage.__class__.__name__} but moving it "
                                f"to its permanent location failed. File has not been moved and record has temporary key still set. "
                                f"See record {instance.__class__.__name__}(pk={instance.pk}) and AsyncTempFile(pk={temp_file.pk}). "
                                f"AsyncTempFile will not be deleted.",
                                exc_info=True,
                            )
                            import traceback

                            temp_file.error = traceback.format_exc()
                            temp_file.save()
                        else:
                            field.before_move_file(instance)
                            setattr(instance, field.name, file)
                            should_save = True
            if not should_save:
                return

            self._save_in_progress.append(instance)
            try:
                instance.save()
                AsyncTempFile.objects.bulk_update(pending_save, ["moved_to_location"])
            finally:
                self._save_in_progress = list(filter(lambda x: x is not instance, self._save_in_progress))


fields_registry = AsyncFileModelRegistry()

signals.post_save.connect(fields_registry._move_temp_files_into_place, sender=None)

# by default cap to 100MiB
default_max_size = 100
# Default max_length on underlying CharField. Note that changing this after fields exist won't be detected by djangos migration system
default_max_length = 500

_UNSET = object()


class AsyncFileMixin(AsyncFileMixinProtocol):
    """Mixin for file fields that works with :class:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage` to
    handle uploading directly to external service like S3 or Azure.

    This field works in conjunction with :class:`~alliance_platform.storage.async_uploads.views.GenerateUploadUrlView`. The view will
    generate a URL (eg. a signed URL when using S3) that the frontend can then upload to. Each view is tied
    to a specific registry which you can specify in ``async_field_registry`` (defaults to :data:`~alliance_platform.storage.async_uploads.registry.default_async_field_registry`).

    The permissions used by :class:`~alliance_platform.storage.async_uploads.views.GenerateUploadUrlView` can be specified in
    ``perm_create`` and ``perm_update``. If not provided they default to the value returned by :meth:`alliance_platform.auth.resolve_perm_name`
    for the action 'create' and 'update' respectively. To disable checking permissions, you can pass ``None`` - this will
    mean any user, including anonymous users, can generate upload urls.

    When using django forms :class:`~alliance_platform.storage.async_uploads.forms.AsyncFileField` provides a widget for handling
    the upload from the frontend. This is the default ``formfield`` provided by :class:`~alliance_platform.storage.async_uploads.models.AsyncFileField`.

    .. note:: The key for the file is stored in the database as a CharField and as such has a max_length. The default
        for this is ``500``. This must be sufficient to accommodate the temporary file value which looks something like
        ``async-temp-files/2021/03/03/fVy5cSVBQpOb-test.png`` by default (see :meth:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.generate_temporary_path`
        for how to customise this).

        Anything after the ``-`` will be truncated if necessary for the purposes of the temporary key. The actual filename
        will still be used for the final field value after the temporary file is moved to the permanent location. When
        moving to the permanent location the filename may be truncated if the length is too long when ``upload_to`` is
        taken into account.

        If using :class:`~alliance_platform.storage.s3.S3AsyncUploadStorage` you may wish to set ``AWS_S3_FILE_OVERWRITE = False`` to avoid overwriting files
        if the same key is used. This is not necessary if your files generate unique paths (eg. ``upload_to`` takes into
        account the record ID and filename).

    The flow for how this works is as follows:

        1) When a form is rendered on the frontend (eg. using :class:`~alliance_platform.storage.async_uploads.forms.AsyncFileField`) it
           knows the ``async_field_id`` from the registry and ``generate_upload_url`` which is to the :class:`~alliance_platform.storage.async_uploads.views.GenerateUploadUrlView`
           view for the registry. This URL will be used to generated a specific URL for each file to upload to.

        2) When an upload occurs on the frontend it first hits ``generate_upload_url`` and passes the ``async_field_id``, the filename
           and optionally an ``instance_id`` if it's an update for an existing record. :class:`~alliance_platform.storage.async_uploads.views.GenerateUploadUrlView`
           looks up the registry for the ``async_field_id`` to get the field and checks permissions on it (``perm_update`` if
           ``instance_id`` is passed otherwise ``perm_create``). If the permission check passes it will then create a
           :class:`~alliance_platform.storage.async_uploads.models.AsyncTempFile` record to track the ``filename`` passed in and a :meth:`generated key <alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.generate_temporary_path>`
           that will be used to upload the file to a temporary location on the storage backend. The view will return an
           upload url by calling :meth:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.generate_upload_url` along with the generated ``key``.

        3) The frontend will receive the ``upload_url`` and ``key`` and proceed to upload to it. When the form is submitted
           and saved the ``key`` (the value returned from :meth:`generated key <alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.generate_temporary_path>`)
           is the what will be stored in the database.

        4) On save a ``post_save`` signal will be called which will check if the file needs to be moved to a permanent
           location. It does this by calling :meth:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.is_temporary_path` and
           :meth:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.move_file` is called to move the file to its permanent
           location at which point the ``AsyncTempFile`` is deleted.

        5) If an upload occurs but the form isn't submitted :class:`~alliance_platform.storage.async_uploads.models.AsyncTempFile` will be created
           but never deleted. The :class:`~alliance_platform.storage.management.commands.cleanup_async_temp_files` should be run
           periodically to clean these up.

    The url returned when accessing the ``url`` property, eg. ``model.file_field.url`` will always return the URL for
    :class:`~alliance_platform.storage.async_uploads.views.DownloadRedirectView`. This view will check the user has permission to download the
    file by checking the ``perm_download`` permission. If not provided this defaults to the value returned by
    :meth:`alliance_platform.auth.resolve_perm_name` with the action "detail" (ie. if they have permission to view the record
    they can download the file). This view will then redirect to the URL provided by
    :meth:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.generate_download_url`.

    .. note:: You must you this with a storage class that implements :class:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage`.
        Either set DEFAULT_FILE_STORAGE to a class (eg. :class:`~alliance_platform.storage.s3.S3AsyncUploadStorage`) or pass an
        instance in the ``storage`` kwarg.

    For use with a Presto form use :class:`~alliance_platform.storage.async_uploads.rest_framework.AsyncFileField` or :class:`~alliance_platform.storage.async_uploads.rest_framework.AsyncImageField`
    on your serializer. This is handled by default when extending :class:`xenopus_frog_app.base.XenopusFrogAppModelSerializer`.
    Codegen will create this as a ``AsyncFileField`` or ``AsyncImageField`` on the frontend Presto model.
    ``getWidgetForField`` will map these fields to the ``UploadWidget``.

    See :class:`~alliance_platform.storage.async_uploads.models.AsyncFileField` and :class:`~alliance_platform.storage.async_uploads.models.AsyncFileField`
    for Field classes that use this.

    .. warning:: User input should be assigned using the :class:`~alliance_platform.storage.async_uploads.models.AsyncFileInputData` class
        which has validation to disallow things like moving the file by changing the key that passed from the frontend. This
        is handled for you when using :class:`~alliance_platform.storage.async_uploads.models.AsyncFileField`,
        :class:`~alliance_platform.storage.async_uploads.models.AsyncImageField`, :class:`~alliance_platform.storage.async_uploads.rest_framework.AsyncFileField`
        or :class:`~alliance_platform.storage.async_uploads.rest_framework.AsyncImageField`

    If a file cannot be moved for any reason (eg. any exception is raised) then the error will be saved in the
    ``error`` field on the :class:`~alliance_platform.storage.async_uploads.models.AsyncTempFile` record and the temporary key will be retained as
    the field value. The error will also be logged to the ``alliance_platform.storage`` logger. This cannot be resolved automatically
    and so requires manual intervention to cleanup.

    File size can be restricted by passing in ``max_size`` (in MB)

    The file extension/file type can be restricted by passing in ``file_restrictions`` for example:

    .. code-block:: python

        ["image/png"]
        ["image/*"]
        [".doc", "docx"]
        ["application/vnd.openxmlformats-officedocument.wordprocessingml.document"]

    """

    model: type[models.Model]
    storage: AsyncUploadStorage

    def __init__(
        self,
        *args,
        perm_create=_UNSET,
        perm_update=_UNSET,
        perm_detail=_UNSET,
        max_size=default_max_size,
        file_restrictions=None,
        async_field_registry=default_async_field_registry,
        max_length=default_max_length,
        **kwargs,
    ):
        super().__init__(*args, max_length=max_length, **kwargs)
        if not isinstance(self.storage, AsyncUploadStorage):
            raise ValueError(
                "When using AsyncFileMixin the file storage class must extend AsyncUploadStorage. Either set DEFAULT_FILE_STORAGE or pass the 'storage' kwarg"
            )
        self.perm_create = perm_create
        self.perm_update = perm_update
        self.perm_detail = perm_detail
        self.max_size = max_size
        if isinstance(file_restrictions, str):
            file_restrictions = [file_restrictions]
        elif not file_restrictions:
            file_restrictions = []
        self.file_restrictions = file_restrictions
        self.async_field_registry = async_field_registry

    def contribute_to_class(self, cls, name: str, private_only: bool = False):
        super().contribute_to_class(cls, name, private_only=private_only)  # type: ignore[safe-super]
        if not cls._meta.abstract and self.model._meta.app_config:
            field = cast(AsyncFileField, self)
            self.async_field_registry.register_field(field)
            if self.perm_update is _UNSET:
                self.perm_update = resolve_perm_name(
                    self.model,
                    action="update",
                    is_global=False,
                )
            if self.perm_create is _UNSET:
                self.perm_create = resolve_perm_name(
                    self.model,
                    action="create",
                    is_global=True,
                )

            if self.perm_detail is _UNSET:
                self.perm_detail = resolve_perm_name(
                    self.model,
                    action="detail",
                    is_global=False,
                )

    def is_valid_filetype(self, filename):
        if not self.file_restrictions:
            return True
        mimetype, _ = mimetypes.guess_type(filename)
        for restriction in self.file_restrictions:
            # special cases for /*
            if restriction in ("image/*", "video/*", "audio/*"):
                if mimetype and mimetype.startswith(restriction[:-1]):
                    return True
            elif restriction.startswith("."):
                # filename restriction
                if filename.lower().endswith(restriction.lower()):
                    return True
            elif mimetype == restriction:
                return True
        return False

    def before_process_file(self, instance):
        """Hook to do something before file is processed."""
        pass

    def before_move_file(self, instance):
        """Hook to do something before file is moved."""
        pass


class AsyncFileInputDataKwargs(TypedDict, total=False):
    key: str
    name: str
    width: int | None
    height: int | None
    error: str | None


@dataclass
class AsyncFileInputData:
    """The data we receive from the frontend gets converted to an instance of this

    :code:`width` and :code:`height` can optionally be included for images to avoid having
    to calculate dimensions manually when an :class:`~alliance_platform.storage.async_uploads.models.AsyncImageField` uses the
    ``width_field`` and ``height_field`` options.
    """

    # The key the file was uploaded to
    key: str
    # The original name of the file
    name: str
    # The width of the image (optional - only applicable to :class:`~alliance_platform.storage.async_uploads.models.AsyncImageField`)
    width: int | None = None
    # The height of the image (optional - only applicable to :class:`~alliance_platform.storage.async_uploads.models.AsyncImageField`)
    height: int | None = None
    # Only set if there was an upload error on the frontend
    error: str | None = None

    def update_dimension_cache(self, field: models.Field):
        """Updates the dimension_cache on :class:`~alliance_platform.storage.async_uploads.models.AsyncImageField`"""

        width = self.width
        height = self.height
        if width and height:
            # See AsyncImageField.dimension_cache. We do this rather than
            # assigning directly to the width + height fields on instance
            # as when creating a new instance by passing in kwargs then
            # the width + height fields may be set to the the defaults
            # (eg. None) after this runs which would override the values.
            # For example the first assignment would work if we did that
            # but the second where arguments are passed in __init__ may
            # not depending on the order of fields on the model
            # # Works
            # tm = AsyncFileTestModel()
            # tm.image_with_dims = AsyncFileInputData(key="test.png", name="test.png", width=16, height=16)
            # # No luck.. width/height get set to None after image_with_dims __set__ runs
            # tm = AsyncFileTestModel(
            #     image_with_dims=AsyncFileInputData(key="test.png", name="test.png", width=16, height=16)
            # )
            # Instead we cache it on the field and `update_dimension_fields`
            # checks for it otherwise falls back to the default ImageField
            # behaviour.
            setattr(field, "dimension_cache", (width, height))

    @classmethod
    def create_from_user_input(cls, values: dict[str, str | int]):
        """Create instance from user input.

        Throws if invalid keys present in ``values``. ``url`` is accepted but ignored.
        """
        data: AsyncFileInputDataKwargs = {}
        valid_fields = ["key", "name", "width", "height", "error"]
        # We return "url" in format_value but it's not something
        # that is relevant in submission. Remove it here so frontend
        # implementations don't need to worry about it.
        ignore_fields = ["url"]
        invalid_input = []
        for field, value in values.items():
            if field in valid_fields:
                data[field] = values[field]  # type: ignore[literal-required] # mypy issue #7178
            elif field not in ignore_fields:
                invalid_input.append(field)
        if invalid_input:
            raise ValidationError(f"Invalid input keys received: {', '.join(invalid_input)}")
        return cls(**data)

    def validate_key(self, field: AsyncFileField, existing_value: str | AsyncFieldFile | None):
        """Validate the key should be accepted

        This checks to make sure the key is a temporary path or matches the existing field
        value. This prevents manipulation of keys which could result in eg. taking over another records file
        """
        if not field.storage.is_temporary_path(self.key):
            existing_key = existing_value
            if isinstance(existing_value, AsyncFieldFile):
                existing_key = cast(
                    str, existing_value.name
                )  # mypy: seems like it cant infer from property 'name' but IDE can
            if existing_key != self.key:
                logger.error(
                    f"{self.key} was received but is NOT a temporary file path and does not match "
                    f"the current value {existing_value}. This could mean the upload was manipulated "
                    f"to attempt to reference a different records file. This action has been blocked "
                    f"with a validation error."
                )
                raise ValidationError({field.name: "Invalid upload received"})


class FieldFileMixin:
    @property
    def url(self: FieldFileMixinProtocol):
        if not self.name:
            raise ValueError(f"`.url()` called on {self} but {self} is empty")

        # URL downloads all go through a DownloadRedirectView so we can do perm checks
        # See AsyncUploadStorage.generate_download_url for where the actual URL generation occurs.
        url = reverse(self.field.async_field_registry.attached_download_view)
        query = urlencode(
            {
                "field_id": self.field.async_field_registry.generate_id(self.field),
                "instance_id": self.instance.pk,
                # this is here to cache bust when an image is modified
                # assuming AWS_S3_FILE_OVERWRITE is false then even if you upload a
                # file with the same name into the same field, it will have a different name
                # hash is taken to avoid exposing the name to frontend which otherwise isn't
                # done until download view verifies permissions
                "_": shake_256(self.name.encode()).hexdigest(8),
            }
        )
        return f"{url}?{query}"


class AsyncFieldFile(FieldFileMixin, FieldFile):
    pass


class AsyncTempFile(models.Model):
    """Model to track files that are being uploaded to a temporary location

    :class:`~alliance_platform.storage.async_uploads.views.GenerateUploadUrlView` is used to generate a URL to directly upload
    a file to. When this URL is generated an :class:`AsyncTempFile` is created to track the new
    key that is used (eg. /temp/2020/01/04/abc123-myfile.png), the original filename (eg. myfile.png)
    and the specific field it came from (this is done via :class:`alliance_platform.storage.async_uploads.registry.AsyncFieldRegistry`).

    Once a file has been uploaded and the form saved the key recorded here will be saved against the
    underlying file field (either :class:`~alliance_platform.storage.async_uploads.models.AsyncFileField` or :class:`~alliance_platform.storage.async_uploads.models.AsyncImageField`)
    on the target model which will check if that key is a temporary file using an :meth:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.is_temporary_path`. If so the
    file will be moved to its permanent location using :meth:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.move_file`
    and the :class:`AsyncTempFile` record will have the ``moved_to_location`` value set. See :class:`~alliance_platform.storage.async_uploads.models.AsyncFileMixin`
    for where this happens and more details.

    You must run the :class:`~alliance_platform.storage.management.commands.cleanup_async_temp_files` command periodically to
    cleanup this table. This handles two cases: the success case where ``moved_to_location`` is set but the record is
    being kept around for a while to detect duplicate submissions, and the other case where upload occurred on the frontend
    but the form was never submitted and the file was never moved.
    """

    created_at = models.DateTimeField(auto_now_add=True)
    original_filename = models.TextField()
    # This is the temporary location the file will be saved to and is generated using AsyncUploadStorage.generate_temporary_path(filename)
    key = models.CharField(max_length=500)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    field_name = models.CharField(max_length=100)
    # This will be set once the file has been moved to its permanent location. We store this so we can detect duplicate submissions and handle them appropriately.
    moved_to_location = models.TextField(blank=True)
    # This will be set if there was an error trying to move the file. When this is set cleanup_async_temp_files will not remove the record.
    error = models.TextField(null=True)

    @classmethod
    def create_for_field(cls, field: AsyncFileField | AsyncImageField, filename: str) -> AsyncTempFile:
        return AsyncTempFile.objects.create(
            original_filename=filename,
            content_type=ContentType.objects.get_for_model(field.model),
            field_name=field.name,
            key=field.storage.generate_temporary_path(filename, max_length=field.max_length),
        )

    class Meta:
        db_table = "alliance_platform_storage_async_temp_file"


class AsyncFileDescriptor(FileDescriptor):
    """Extracts the ``key`` from AsyncFileInputData to be set against instance"""

    def __set__(self, instance, value):
        if isinstance(value, AsyncFileInputData):
            value.validate_key(cast(AsyncFileField, self.field), instance.__dict__.get(self.field.name))
            value = value.key
        super().__set__(instance, value)


class AsyncFileField(AsyncFileMixin, FileField):  # type: ignore[misc] # mypy dont like the fact that we're using AsyncUploadStorage and report it as conflict w/ FileField
    """A FileField that works with AsyncUploadStorage to directly upload somewhere (eg. S3) from the frontend

    See :class:`~alliance_platform.storage.async_uploads.models.AsyncFileMixin` for more details on how this field works.
    """

    attr_class = AsyncFieldFile
    descriptor_class = AsyncFileDescriptor

    def formfield(self, **kwargs):
        from alliance_platform.storage.async_uploads.forms import AsyncFileField as AsyncFileFormField

        return super().formfield(
            **{
                "form_class": AsyncFileFormField,
                "async_field_id": self.async_field_registry.generate_id(self),
                "async_field_registry": self.async_field_registry,
                "storage": self.storage,
                **kwargs,
            }
        )


class AsyncImageDescriptor(FileDescriptor):
    """Extracts the ``key`` from AsyncFileInputData to be set against instance & handles width/height

    If width & height are present they are set on the field and handled by
    :meth:`alliance_platform.storage.async_uploads.models.AsyncImageField.update_dimension_fields`.
    """

    def __set__(self, instance, value):
        if isinstance(value, AsyncFileInputData):
            value.validate_key(cast(AsyncFileField, self.field), instance.__dict__.get(self.field.attname))
            value.update_dimension_cache(self.field)
            value = value.key
            super().__set__(instance, value)
        else:
            # From ImageFileDescriptor:
            # To prevent recalculating image dimensions when we are instantiating
            # an object from the database (bug #11084), only update dimensions if
            # the field had a value before this assignment.  Since the default
            # value for FileField subclasses is an instance of field.attr_class,
            # previous_file will only be None when we are called from
            # Model.__init__().  The ImageField.update_dimension_fields method
            # hooked up to the post_init signal handles the Model.__init__() cases.
            # Assignment happening outside of Model.__init__() will trigger the
            # update right here.

            # We only do this if we don't have an AsyncFileInputData as
            # that contains the width / height submitted from the
            # frontend and so avoids potentially downloading the file
            # to get the width/height.
            previous_file = instance.__dict__.get(self.field.attname)
            super().__set__(instance, value)
            # Don't compare value != previous_file as it doesn't guarantee the underlying
            # file hasn't changed (eg. the key could be set to same filename again after
            # a new upload but be a different file)
            field = cast(AsyncImageField, self.field)
            if previous_file is not None:
                field.update_dimension_fields(instance, force=True)


class AsyncImageFieldFile(FieldFileMixin, ImageFieldFile):
    pass


class AsyncImageField(AsyncFileMixin, ImageField):  # type: ignore[misc] # mypy dont like the fact that we're using AsyncUploadStorage and report it as conflict w/ FileField
    """A ImageField that works with AsyncUploadStorage to directly upload somewhere (eg. S3) from the frontend

    This supports ``width_field`` and ``height_field`` in two ways

    1) (preferred) The frontend passes the width & height with the form submission. This is supported with the
    :class:`~alliance_platform.storage.async_uploads.forms.AsyncImageField` form field (the default). This works in
    conjunction with the ``UploadWidget`` on the frontend.

    2) (slow) Same behaviour as :class:`~django.db.models.ImageField` which requires the file to be downloaded
    and processed with Pillow.

    See :class:`~alliance_platform.storage.async_uploads.models.AsyncFileMixin` for more details on how this field works.
    """

    descriptor_class = AsyncImageDescriptor
    attr_class = AsyncImageFieldFile

    # Cache dimensions as (width, height) from AsyncImageDescriptor. Used in update_dimension_fields
    dimension_cache: tuple[int, int] | None

    # Don't check if Pillow is installed. This is used to get dimensions of images.
    _supress_pillow_check: bool

    width_field: str | None
    height_field: str | None

    def __init__(self, *args, suppress_pillow_check=False, **kwargs):
        super().__init__(*args, **kwargs)
        self._suppress_pillow_check = suppress_pillow_check
        self.dimension_cache = None

    def check(self, **kwargs):
        if self._suppress_pillow_check:
            return super(ImageField, self).check(**kwargs)
        return super().check(**kwargs)

    def update_dimension_fields(self, instance, force=False, *args, **kwargs):
        """
        NOTE: The base Django image field connects this function to the model's
        ``post_init`` signal in the field's ``contribute_to_class`` function, to
        ensure that dimensions are calculated on model initialisation. However,
        the signal will not be triggered for child models, meaning this function
        will not be called on initialisation when inherited. This should only affect
        the setting of ``width_field`` and ``height_field`` on the model, so most
        functionality of this field should be unaffected. However, it's something to
        keep in mind when using this field on a model that inherits from a non-abstract
        model.

        If we have a cache use that otherwise fall back to default. Note that
        the default will open the file to work out dimensions and so with
        remote backends like S3 would result in downloading the file first.
        """
        if self.dimension_cache:
            has_dimension_fields = self.width_field or self.height_field
            if not has_dimension_fields or self.attname not in instance.__dict__:
                return
            width, height = self.dimension_cache
            self.dimension_cache = None
            if self.width_field:
                setattr(instance, self.width_field, width)
            if self.height_field:
                setattr(instance, self.height_field, height)
            return width, height

        else:
            super().update_dimension_fields(instance, force, *args, **kwargs)

    @property
    def expects_dimension_cache(self):
        return (self.width_field and self.height_field) and (not self.dimension_cache)

    def before_process_file(self, instance):
        """
        The base ImageField ensures that ``update_dimension_fields`` is called on initialisation using ``contribute_to_class``
        and the ``post_init`` signal. Those can't be relied on for child classes - so here we double-check that width and
        height are set when we expect them to be set, and if not, manually run ``update_dimension_fields``
        """
        if self.expects_dimension_cache:
            width = getattr(instance, self.width_field)
            height = getattr(instance, self.height_field)
            if (width is None) or (height is None):
                self.update_dimension_fields(instance, force=True)

    def before_move_file(self, instance):
        """Prior to moving file populate the dimensions cache

        This is so the moved file retains same dimensions as prior to move without
        needing to recalculate it (ie. it's the same file, just in a different location).
        """
        if self.expects_dimension_cache:
            width = getattr(instance, cast(str, self.width_field))
            height = getattr(instance, cast(str, self.height_field))
            self.dimension_cache = (width, height)

    def formfield(self, **kwargs):
        from alliance_platform.storage.async_uploads.forms import AsyncImageField as AsyncImageFormField

        return super().formfield(
            **{
                "form_class": AsyncImageFormField,
                "async_field_id": self.async_field_registry.generate_id(self),
                "async_field_registry": self.async_field_registry,
                "storage": self.storage,
                **kwargs,
            }
        )



from typing import Any
from typing import TypedDict

from alliance_platform.base_settings import AlliancePlatformSettingsBase


class AlliancePlatformStorageSettingsType(TypedDict, total=False):
    """Settings for the Storage package of the Alliance Platform

    These can be set in the Django settings file under the ``ALLIANCE_PLATFORM`` key:

    .. code-block:: python

        ALLIANCE_PLATFORM = {
            "STORAGE": {
                # settings go here
            }
        }
    """

    #: Number of seconds a generated upload URL is valid for by default. Defaults to 3600.
    UPLOAD_URL_EXPIRY: int | None
    #: Number of seconds a generated download URL is valid for by default. Defaults to 3600.
    DOWNLOAD_URL_EXPIRY: int | None


class AlliancePlatformStorageSettings(AlliancePlatformSettingsBase):
    #: Number of seconds a generated upload URL is valid for by default. Defaults to 3600.
    UPLOAD_URL_EXPIRY: int
    #: Number of seconds a generated download URL is valid for by default. Defaults to 3600.
    DOWNLOAD_URL_EXPIRY: int


DEFAULTS: Any = {
    "UPLOAD_URL_EXPIRY": 3600,
    "DOWNLOAD_URL_EXPIRY": 3600,
}

ap_storage_settings = AlliancePlatformStorageSettings(
    "STORAGE",
    defaults=DEFAULTS,
)



from typing import Any
from typing import cast

from alliance_platform.storage.async_uploads.registry import default_async_field_registry
from alliance_platform.storage.async_uploads.storage.filesystem import FileSystemAsyncUploadStorage
from alliance_platform.storage.async_uploads.views.api import ViewProtocol
from alliance_platform.storage.settings import ap_storage_settings
from django.core.files.uploadedfile import UploadedFile
from django.core.signing import BadSignature
from django.core.signing import SignatureExpired
from django.http import FileResponse
from django.http import HttpRequest
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.http import HttpResponseForbidden
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import csrf_exempt


@method_decorator(csrf_exempt, name="dispatch")
class FileSystemAsyncStorageUploadView(View):
    """View to handle a direct upload to the filesystem when using :class:`~alliance_platform.storage.async_uploads.storage.filesystem.FileSystemAsyncUploadStorage`.

    Each view is tied to a specific registry which you can specify in the ``registry`` kwarg (defaults to :data:`~alliance_platform.storage.async_uploads.registry.default_async_field_registry`).

    Note that this view does works off of signed data generated by :class:`~alliance_platform.storage.async_uploads.storage.filesystem.FileSystemAsyncUploadStorage.generate_upload_url`,
    and does not itself check any permissions. It does verify the signed data, including enforcing an expiry based on
    :data:`~alliance_platform.storage.settings.AlliancePlatformStorageSettings.UPLOAD_URL_EXPIRY`. This is similar to a
    signed URL generated by the S3 or Azure backends.

    The permission checking is done in :class:`~alliance_platform.storage.async_uploads.views.GenerateUploadUrlView`, which
    calls :class:`~alliance_platform.storage.async_uploads.storage.filesystem.FileSystemAsyncUploadStorage.generate_upload_url`.

    The URL for this view is automatically registered when :class:`~alliance_platform.storage.async_uploads.storage.filesystem.FileSystemAsyncUploadStorage`
    is used if you have followed the :ref:`register-urls` guide.
    """

    registry = default_async_field_registry

    registration_attach_key = "file_system_async_upload_storage_upload_view"

    @classmethod
    def as_view(cls, **initkwargs: Any):
        view: ViewProtocol = cast(ViewProtocol, super().as_view(**initkwargs))
        registry = view.view_initkwargs.get("registry", default_async_field_registry)
        if hasattr(registry, cls.registration_attach_key):
            import warnings

            warnings.warn(
                f"{registry} has already been attached to one FileSystemAsyncUploadStorage. Pass a different registry in like `FileSystemAsyncStorageUploadView.as_view({{ registry: another_reg }})`"
            )
        setattr(registry, cls.registration_attach_key, view)
        return view

    def post(self, request: HttpRequest):
        signed_path = request.GET.get("path")
        field_id = request.GET.get("field_id")

        if not signed_path or not field_id:
            return HttpResponseBadRequest("Missing signed path.")
        field = self.registry.fields_by_id.get(field_id)
        if not field:
            return HttpResponseBadRequest("Missing field_id.")
        storage = cast(FileSystemAsyncUploadStorage, field.storage)

        try:
            path = storage.signer.unsign(signed_path, max_age=ap_storage_settings.UPLOAD_URL_EXPIRY)
        except SignatureExpired:
            return HttpResponseForbidden("Upload URL has expired.")
        except BadSignature:
            return HttpResponseForbidden("Invalid upload URL.")

        if "file" not in request.FILES:
            return HttpResponseBadRequest("No file uploaded.")

        # Save the uploaded file to the specified path
        uploaded_file = cast(UploadedFile, request.FILES["file"])
        storage.save(path, uploaded_file)

        return HttpResponse("File uploaded successfully.", status=201)


class FileSystemAsyncStorageDownloadView(View):
    """View to handle a direct download from the filesystem when using :class:`~alliance_platform.storage.async_uploads.storage.filesystem.FileSystemAsyncUploadStorage`.

    Each view is tied to a specific registry which you can specify in the ``registry`` kwarg (defaults to :data:`~alliance_platform.storage.async_uploads.registry.default_async_field_registry`).

    Note that this view does works off of signed data generated by :class:`~alliance_platform.storage.async_uploads.storage.filesystem.FileSystemAsyncUploadStorage.generate_download_url`,
    and does not itself check any permissions. It does verify the signed data, including enforcing an expiry based on
    :data:`~alliance_platform.storage.settings.AlliancePlatformStorageSettings.DOWNLOAD_URL_EXPIRY`. This is similar to a
    signed URL generated by the S3 or Azure backends.

    The permission checking is done in :class:`~alliance_platform.storage.async_uploads.views.GenerateUploadUrlView`, which
    calls :class:`~alliance_platform.storage.async_uploads.storage.filesystem.FileSystemAsyncUploadStorage.generate_upload_url`.

    The URL for this view is automatically registered when :class:`~alliance_platform.storage.async_uploads.storage.filesystem.FileSystemAsyncUploadStorage`
    is used if you have followed the :ref:`register-urls` guide.
    """

    registry = default_async_field_registry

    registration_attach_key = "file_system_async_upload_storage_download_view"

    @classmethod
    def as_view(cls, **initkwargs: Any):
        view: ViewProtocol = cast(ViewProtocol, super().as_view(**initkwargs))
        registry = view.view_initkwargs.get("registry", default_async_field_registry)
        if hasattr(registry, cls.registration_attach_key):
            import warnings

            warnings.warn(
                f"{registry} has already been attached to one FileSystemAsyncDownloadStorage. Pass a different registry in like `FileSystemAsyncStorageDownloadView.as_view({{ registry: another_reg }})`"
            )
        setattr(registry, cls.registration_attach_key, view)
        return view

    def get(self, request: HttpRequest):
        signed_path = request.GET.get("path")
        field_id = request.GET.get("field_id")

        if not signed_path or not field_id:
            return HttpResponseBadRequest("Missing signed path.")
        field = self.registry.fields_by_id.get(field_id)
        if not field:
            return HttpResponseBadRequest("Missing field_id.")
        storage = cast(FileSystemAsyncUploadStorage, field.storage)

        try:
            path = storage.signer.unsign(signed_path, max_age=ap_storage_settings.DOWNLOAD_URL_EXPIRY)
        except SignatureExpired:
            return HttpResponseForbidden("Download URL has expired.")
        except BadSignature:
            return HttpResponseForbidden("Invalid download URL.")

        file_handle = storage.open(path, "rb")
        filename = path.split("/")[-1]
        response = FileResponse(file_handle, as_attachment=True, filename=filename)
        return response



import logging
from unittest import mock

import alliance_platform.storage.async_uploads.models as async_file_fields
from alliance_platform.storage.async_uploads.models import AsyncTempFile
from alliance_platform.storage.async_uploads.models import default_max_length as async_file_max_length
from alliance_platform.storage.async_uploads.rest_framework import AsyncFileField
from alliance_platform.storage.async_uploads.rest_framework import AsyncImageField
from django.test import TestCase
from rest_framework import serializers
from test_alliance_platform_storage.models import AsyncFileTestModel
from test_alliance_platform_storage.models import AsyncFileTestParentModel


class AsyncFileTestModelSerializer(serializers.ModelSerializer):
    serializer_field_mapping = {
        **serializers.ModelSerializer.serializer_field_mapping,
        async_file_fields.AsyncFileField: AsyncFileField,
        async_file_fields.AsyncImageField: AsyncImageField,
    }

    class Meta:
        fields = ("file1", "image_with_dims")
        model = AsyncFileTestModel


class AsyncFileTestModelParentSerializer(serializers.ModelSerializer):
    files = AsyncFileTestModelSerializer(many=True, required=False)

    class Meta:
        fields = ("files",)
        model = AsyncFileTestParentModel


class SerializerAsyncFileTestCase(TestCase):
    def test_file_field(self):
        pdf = AsyncTempFile.create_for_field(AsyncFileTestModel._meta.get_field("file1"), "test.pdf")
        image = AsyncTempFile.create_for_field(
            AsyncFileTestModel._meta.get_field("image_with_dims"), "test.png"
        )
        serializer = AsyncFileTestModelSerializer(
            data=dict(
                file1={"key": pdf.key, "name": pdf.original_filename},
                image_with_dims={
                    "key": image.key,
                    "name": image.original_filename,
                    "width": 16,
                    "height": 24,
                },
            )
        )
        serializer.is_valid(raise_exception=True)
        with mock.patch("test_alliance_platform_storage.storage.DummyStorage._open") as mock_method:
            instance = serializer.save()

            mock_method.assert_not_called()

            self.assertEqual(instance.file1, "test.pdf")
            self.assertEqual(instance.image_with_dims, "test.png")
            self.assertEqual(instance.image_width, 16)
            self.assertEqual(instance.image_height, 24)

    def test_reject_bad_input(self):
        # Can't change the key on a field
        serializer = AsyncFileTestModelSerializer(data={"image_with_dims": "whatever.png"})
        self.assertFalse(serializer.is_valid())
        self.assertRegex(serializer.errors["image_with_dims"][0], "Bad input for file field")

        with self.assertLogs(logging.getLogger("alliance_platform.storage")):
            # Key isn't prefixed as a temporary path so must be rejected
            serializer = AsyncFileTestModelSerializer(
                data={"image_with_dims": {"key": "/storage/test.png", "name": "test.png"}}
            )
            self.assertFalse(serializer.is_valid())
            self.assertRegex(serializer.errors["image_with_dims"][0], "Invalid upload received")

        # Check max_length is validated
        tf = AsyncTempFile.create_for_field(
            AsyncFileTestModel._meta.get_field("image_with_dims"), "a" * (async_file_max_length + 1)
        )
        serializer = AsyncFileTestModelSerializer(
            data={"image_with_dims": {"key": tf.key, "name": tf.original_filename}}
        )
        self.assertFalse(serializer.is_valid())

        tf = AsyncTempFile.create_for_field(
            AsyncFileTestModel._meta.get_field("image_with_dims"), "a" * async_file_max_length
        )
        serializer = AsyncFileTestModelSerializer(
            data={"image_with_dims": {"key": tf.key, "name": tf.original_filename}}
        )
        self.assertTrue(serializer.is_valid())

        tf = AsyncTempFile.create_for_field(AsyncFileTestModel._meta.get_field("image_with_dims"), "test.png")
        tm = AsyncFileTestModel.objects.create(
            image_with_dims=async_file_fields.AsyncFileInputData(
                key=tf.key, name=tf.original_filename, width=32, height=32
            )
        )
        self.assertEqual(tm.image_with_dims.name, "test.png")
        self.assertEqual(tm.image_width, 32)
        self.assertEqual(tm.image_height, 32)

        # Key matches existing record, should be fine
        serializer = AsyncFileTestModelSerializer(
            data={"image_with_dims": {"key": "test.png", "name": "test.png"}}, instance=tm
        )
        self.assertTrue(serializer.is_valid())

        with self.assertLogs(logging.getLogger("alliance_platform.storage")):
            # Changing the key on existing item isn't accepted (unless it's a temp path)
            serializer = AsyncFileTestModelSerializer(
                data={"image_with_dims": {"key": "test2.png", "name": "test.png"}}, instance=tm
            )
            self.assertFalse(serializer.is_valid())
            self.assertRegex(serializer.errors["image_with_dims"][0], "Invalid upload received")

        tf = AsyncTempFile.create_for_field(
            AsyncFileTestModel._meta.get_field("image_with_dims"), "test2.png"
        )
        serializer = AsyncFileTestModelSerializer(
            data={"image_with_dims": {"key": tf.key, "name": "test2.png", "width": 16, "height": 16}},
            instance=tm,
        )
        self.assertTrue(serializer.is_valid())
        serializer.save()
        tm = AsyncFileTestModel.objects.get(pk=tm.pk)

        self.assertEqual(tm.image_with_dims.name, "test2.png")
        self.assertEqual(tm.image_width, 16)
        self.assertEqual(tm.image_height, 16)

    def test_nested_serializer_behavior(self):
        parent = AsyncFileTestParentModel.objects.create()
        temp_file = AsyncTempFile.create_for_field(
            AsyncFileTestModel._meta.get_field("image_with_dims"), "test.png"
        )
        test_model = AsyncFileTestModel.objects.create(
            parent=parent,
            image_with_dims=async_file_fields.AsyncFileInputData(
                key=temp_file.key, name=temp_file.original_filename, width=32, height=32
            ),
        )

        # We have a workaround for nested serializers to track the instance, which is otherwise unavailable in async_uploads.rest_framework.
        # This test covers that patch - if this fails in the future it likely indicates async_uploads.rest_framework has better
        # handling of nested serializers and we can drop our patch.
        with self.assertLogs(logging.getLogger("alliance_platform.storage")):
            data = AsyncFileTestModelSerializer(instance=test_model).data
            nesting_serializer = AsyncFileTestModelParentSerializer(
                data={"files": [data]},
                instance=parent,
            )
            self.assertFalse(nesting_serializer.is_valid())

        # emulates FE payload where id would be included
        data["id"] = test_model.id

        # test that when frontend submits a nested request with unchanged data, it succeeds
        nesting_serializer = AsyncFileTestModelParentSerializer(
            data={"files": [data]},
            instance=parent,
        )
        self.assertTrue(nesting_serializer.is_valid())

        # test that changing the file also wont raise a validation error on changed key
        new_temp_file = AsyncTempFile.create_for_field(
            AsyncFileTestModel._meta.get_field("image_with_dims"), "test2.png"
        )
        data["image_with_dims"] = {"key": new_temp_file.key, "name": "test2.png", "width": 16, "height": 16}
        nesting_serializer = AsyncFileTestModelParentSerializer(
            data={"files": [data]},
            instance=parent,
        )
        self.assertTrue(nesting_serializer.is_valid())



from alliance_platform.storage.settings import ap_storage_settings
from django.apps.config import AppConfig


class AlliancePlatformStorageConfig(AppConfig):
    name = "alliance_platform.storage"
    verbose_name = "Alliance Platform Storage"
    label = "alliance_platform_storage"

    def ready(self):
        ap_storage_settings.check_settings()



from __future__ import annotations

from hashlib import shake_256
from typing import TYPE_CHECKING
from typing import Callable

from alliance_platform.storage.async_uploads.storage.base import AsyncUploadStorage
from django.db.models import Field
from django.http import HttpResponse

if TYPE_CHECKING:
    # circular imports
    from alliance_platform.storage.async_uploads.models import AsyncFileField
    from alliance_platform.storage.async_uploads.models import AsyncImageField


def is_same_field(field1: Field, field2: Field) -> bool:
    """Are two fields the same?

    This doesn't just do equality check as in some cases (eg. test runner) the
    same model appears to be loaded twice
    """
    return field1.model._meta.label == field2.model._meta.label and field1.name == field2.name


class AsyncFieldRegistry:
    """A registry for async fields. This allows looking up a field by its ID.

    In most cases the default registry :code:`default_async_field_registry` is suitable and more than one
    registry is not necessary.
    """

    # This is for easy lookup using the unique name for a class. The name is what
    # can be passed from the frontend to identify a particular registration
    fields_by_id: dict[str, AsyncFileField | AsyncImageField]
    # The attached GenerateUploadUrlView as returned by GenerateUploadUrlView.as_view(). This is attached automatically
    # when GenerateUploadUrlView.as_view() is called.
    attached_view: Callable[..., HttpResponse] | None = None

    # The attached AsyncFileDownloadView as returned by AsyncFileDownloadView.as_view(). This is attached automatically
    # when AsyncFileDownloadView.as_view() is called.
    attached_download_view: Callable[..., HttpResponse] | None = None

    def __init__(self, name):
        """
        Args:
            name: A name for this registration. This is just used to make debugging easier when printing out registration instances.
        """
        self.name = name
        self.fields_by_id = {}

    def __str__(self):
        return f"AsyncFieldRegistry(name='{self.name}')"

    def generate_id(self, field: Field):
        """Generate a string ID for a field to be used to lookup into ``fields_by_id``.

        This is passed from the frontend to uniquely identify a field in a registry.
        """
        # Don't expose label to the frontend
        label = shake_256(field.model._meta.label.encode()).hexdigest(8)
        field_name = shake_256(field.name.encode()).hexdigest(8)
        # Include field name to make debugging a bit easier and avoid collisions. This isn't
        # exposing anything we care about as this ID is always associated with the named field
        # on the frontend anyway
        return f"{label}.{field_name}.{field.name}"

    def register_field(self, field: AsyncFileField | AsyncImageField):
        """Register a field in this registry."""
        if not isinstance(getattr(field, "storage", None), AsyncUploadStorage):
            raise ValueError(
                f"Only fields that implement AsyncUploadStorage are supported. {field} does not implement AsyncUploadStorage."
            )
        field_id = self.generate_id(field)
        if field_id in self.fields_by_id:
            if not is_same_field(self.fields_by_id[field_id], field):
                raise ValueError(
                    f"Collision on field id '{field_id}' for field {field} vs {self.fields_by_id[field_id]}"
                )
        else:
            # don't want to register fake migration models
            if field.model.__module__ == "__fake__":
                return

            self.fields_by_id[field_id] = field

    def get_url_patterns(self):
        """Get the required URL patterns for the storage classes used by the registered fields

        This calls :meth:`~alliance_platform.storage.async_uploads.base.AsyncUploadStorage.get_url_patterns` on each
        storage instance and combines the results.

        Usage::

            from django.urls import path
            from django.urls import include
            from alliance_platform.storage.async_uploads.registry import default_async_field_registry

            urlpatterns = [path("async-uploads/", include(default_async_field_registry.get_url_patterns()))]
        """
        storage_instances = {field.storage for field in self.fields_by_id.values()}

        patterns = []
        for storage in storage_instances:
            patterns += storage.get_url_patterns(self)
        return patterns


#: The default registry class to use when none is provided.
default_async_field_registry = AsyncFieldRegistry("default")



import json
from json import JSONDecodeError

from alliance_platform.storage.async_uploads.models import AsyncFileInputData
from alliance_platform.storage.async_uploads.models import AsyncTempFile
from alliance_platform.storage.async_uploads.registry import AsyncFieldRegistry
from alliance_platform.storage.async_uploads.storage.base import AsyncUploadStorage
from django import forms
from django.core import validators
from django.core.exceptions import ValidationError
from django.forms.widgets import Input
from django.urls import reverse
from django.utils.deconstruct import deconstructible


# This does not extend FileInput as we don't actually receive a file
# from the frontend. Instead we get a string which represents the key
# for the storage backend.
class AsyncFileInput(Input):
    """Input for handling async uploads

    This handles the submission value from UploadWidget and converts it to an instance
    of :class:`~alliance_platform.storage.async_uploads.models.AsyncFileInputData`. This is then handled
    on the descriptor classes for :class:`~alliance_platform.storage.async_uploads.models.AsyncFileField`
    and :class:`~alliance_platform.storage.async_uploads.models.AsyncImageField`.

    To customise the widget rendered on the frontend you can override the template
    ``alliance_platform/storage/widgets/async_file_input.html``.
    """

    input_type = "async-file"
    template_name = "alliance_platform/storage/widgets/async_file_input.html"

    # This is needed to generate URL to file in case where we only
    # have the AsyncTempFile (see `format_value`)
    storage: AsyncUploadStorage

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)

        # Resolving the upload url needs to be delayed until render so that the
        # GenerateUploadUrl view has a chance to attach to the registry. Resolve
        # that here.
        async_field_registry = context["widget"]["attrs"].pop("async_field_registry", None)
        if not async_field_registry:
            raise ValueError("AsyncFileInput expects a 'async_field_registry' to be passed in widget attrs")

        context["widget"]["attrs"]["generate_upload_url"] = reverse(async_field_registry.attached_view)

        return context

    def format_value(self, value):
        """Given a value return it in serialized format expected by frontend

        Frontend expects a json string like:

            {"key": "/some/storage/location/test.png", "name": "test.png" }
        """
        if not value:
            return None
        if isinstance(value, AsyncFileInputData):
            # If a submission occurs but isn't successful (eg. validation fails on another
            # field) we end up with a string which should match the key on AsyncTempFile
            try:
                temp_file = AsyncTempFile.objects.get(key=value.key)
                return json.dumps(
                    {
                        "key": temp_file.key,
                        "name": temp_file.original_filename,
                        "url": self.storage.url(temp_file.key),
                    }
                )
            except AsyncTempFile.DoesNotExist:
                return json.dumps(
                    {
                        "key": value.key,
                        "name": value.name,
                    }
                )
        try:
            url = value.url
        except Exception:
            # url isn't critical; if it fails ignore it and frontend may just not be able
            # to show a thumbnail
            url = None
        return json.dumps({"key": value.name, "name": value.name.split("/")[-1], "url": url})

    def value_from_datadict(self, data, files, name):
        """This expects to receive a valid json string that can be used to instantiate AsyncFileInputData

        If invalid JSON is received this is handled by the ``AsyncFileInputDataValidator`` which will
        raise a ValidationError. We load the json here rather than in the field ``clean`` so that we have
        the ``AsyncFileInputData`` in ``format`` for the case where there's a ValidationError and form
        re-renders.

        AsyncFileDescriptor & AsyncImageDescriptor handle extracting the ``key`` from this which is
        what is set against the field on the model.

        We need the AsyncFileInputData to pass across extra details that we may need - eg. width & height
        for image fields
        """
        value = data.get(name)
        try:
            value = json.loads(value)
            if value:
                return AsyncFileInputData.create_from_user_input(value)
            return value
        except JSONDecodeError:
            return value


class AsyncFileInputDataLengthValidator(validators.MaxLengthValidator):
    """Validate length of ``key`` on an AsyncFileInputData

    FileField passes ``max_length`` to the form field and the field is, at the database level,
    a char field. The normal FileField handles this internally but we can't extend forms.FileField
    as it expects to be handling File submissions which we aren't doing here. This validator
    just extracts the ``key`` and validates the length of that which is what gets written to the
    database.
    """

    def clean(self, x):
        if isinstance(x, AsyncFileInputData):
            # We want to check the length of the key which is what gets stored in db
            return len(x.name)
        return len(x)


@deconstructible
class AsyncFileInputDataValidator:
    """AsyncFileInputData has an error key that can be sent from the frontend - this validator just checks that"""

    def __call__(self, value):
        if isinstance(value, AsyncFileInputData):
            if value.error:
                raise ValidationError(f"There was an unexpected error: {value.error}")
        else:
            raise ValidationError("Bad input for file field")


class AsyncFileField(forms.Field):
    """Form field that renders a :class:`~alliance_platform.storage.async_uploads.models.AsyncFileInput`

    This is the default form field for :class:`alliance_platform.storage.async_uploads.models.AsyncFileField`
    """

    widget = AsyncFileInput

    async_field_registry: AsyncFieldRegistry
    async_field_id: str

    def __init__(
        self,
        *args,
        async_field_registry: AsyncFieldRegistry,
        async_field_id: str,
        storage: AsyncUploadStorage,
        max_length=None,
        **kwargs,
    ):
        """

        Args:
            *args: Any additional arguments to pass through to :class:`django.forms.Field`
            async_field_registry: The async field registry that should used on the frontend to create
                unique upload urls for files. This typically comes from :code:`field.async_field_registry` where :code:`field`
                is an :class:`~alliance_platform.storage.async_uploads.models.AsyncFileField`.
            async_field_id: The string ID of the field used in the registry. Generated with ``field.async_field_registry.generate_id(field)``.
            storage: The storage class. This comes from the field ``field.storage``.
            max_length: Max length for the filename
            **kwargs:ny additional keyword arguments to pass through to :class:`django.forms.Field`
        """
        self.async_field_registry = async_field_registry
        self.async_field_id = async_field_id
        # This comes from FileField
        self.max_length = max_length
        super().__init__(*args, **kwargs)

        # No way to pass args to a widget with how forms.Field work.. just
        # set it after creation
        self.widget.storage = storage

        if max_length is not None:
            self.validators.append(AsyncFileInputDataLengthValidator(int(max_length)))
        self.validators.append(AsyncFileInputDataValidator())

    def widget_attrs(self, widget):
        attrs = super().widget_attrs(widget)
        attrs["async_field_registry"] = self.async_field_registry
        attrs["async_field_id"] = self.async_field_id
        return attrs


class AsyncImageField(AsyncFileField):
    widget = AsyncFileInput

    def widget_attrs(self, widget):
        attrs = super().widget_attrs(widget)
        if "accept" not in attrs:
            attrs.setdefault("accept", "image/*")
        if "list_type" not in attrs:
            attrs.setdefault("list_type", "picture")
        return attrs



from hashlib import shake_256
import json
import logging
import os
from typing import Literal
from typing import cast
from unittest import mock
from unittest.mock import patch
from urllib.parse import urlencode

from alliance_platform.storage.async_uploads.forms import AsyncFileInput
from alliance_platform.storage.async_uploads.models import AsyncFileField
from alliance_platform.storage.async_uploads.models import AsyncFileInputData
from alliance_platform.storage.async_uploads.models import AsyncTempFile
from alliance_platform.storage.async_uploads.models import default_max_length as async_file_max_length
from alliance_platform.storage.async_uploads.registry import default_async_field_registry
from django.core.exceptions import SuspiciousFileOperation
from django.core.exceptions import ValidationError
from django.core.files import File
from django.forms import ModelForm
from django.test import TestCase
from django.test import override_settings
from django.urls import reverse
from test_alliance_platform_storage.models import AlternateRegistryModel
from test_alliance_platform_storage.models import AsyncFilePermTestModel
from test_alliance_platform_storage.models import AsyncFileTestChildModel
from test_alliance_platform_storage.models import AsyncFileTestModel
from test_alliance_platform_storage.models import User
from test_alliance_platform_storage.models import another_registry
from test_alliance_platform_storage.storage import DummyStorage

TEST_IMAGE_PATH = os.path.join(
    os.path.dirname(__file__) + "/../test_alliance_platform_storage/files/test.png"
)
TEST_IMAGE_64_64_PATH = os.path.join(
    os.path.dirname(__file__) + "/../test_alliance_platform_storage/files/test-64x64.png"
)

DUMMY_STORAGES_SETTING = {
    "default": {
        "BACKEND": "test_alliance_platform_storage.storage.DummyStorage",
    },
    "staticfiles": {
        "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage",
    },
}


class AsyncFileMixinTestCase(TestCase):
    MODELS_TO_TEST = [AsyncFileTestModel, AsyncFileTestChildModel]

    def test_validate_storage_class(self):
        with override_settings(
            STORAGES={
                "default": {
                    "BACKEND": "django.core.files.storage.FileSystemStorage",
                },
                "staticfiles": {
                    "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage",
                },
            }
        ):
            with self.assertRaisesRegex(ValueError, "file storage class must extend AsyncUploadStorage"):
                AsyncFileField()
        with override_settings(STORAGES=DUMMY_STORAGES_SETTING):
            AsyncFileField()

    def test_registry(self):
        field1 = AlternateRegistryModel._meta.get_field("file1")
        field2 = AlternateRegistryModel._meta.get_field("file2")
        self.assertEqual(field1.async_field_registry, default_async_field_registry)
        self.assertEqual(
            default_async_field_registry.fields_by_id[default_async_field_registry.generate_id(field1)],
            field1,
        )
        self.assertEqual(field2.async_field_registry, another_registry)
        self.assertEqual(another_registry.fields_by_id[another_registry.generate_id(field2)], field2)

    def test_image_field_dimensions(self):
        for async_model_to_test in self.MODELS_TO_TEST:
            with self.subTest(async_model_to_test.__name__):
                with mock.patch("test_alliance_platform_storage.storage.DummyStorage._open") as mock_method:
                    tf = AsyncTempFile.create_for_field(
                        async_model_to_test._meta.get_field("image_with_dims"), "example.png"
                    )
                    tm = async_model_to_test.objects.create(
                        image_with_dims=AsyncFileInputData(
                            key=tf.key, name=tf.original_filename, width=32, height=32
                        )
                    )
                    self.assertEqual(tm.image_with_dims, "example.png")
                    self.assertEqual(tm.image_width, 32)
                    self.assertEqual(tm.image_height, 32)

                    # If width & height are set using AsyncFileInputData then storage should not have to open the file
                    mock_method.assert_not_called()

                    mock_method.return_value = File(open(TEST_IMAGE_PATH, "rb"))

                    # This should result in storage opening the file and reading dimensions from there
                    tm.image_with_dims = "test.png"
                    tm.save()

                    self.assertEqual(tm.image_with_dims, "test.png")
                    self.assertEqual(tm.image_width, 16)
                    self.assertEqual(tm.image_height, 16)

                    mock_method.assert_called_once_with("test.png", "rb")
                    mock_method.mock_reset()

                with mock.patch("test_alliance_platform_storage.storage.DummyStorage._open") as mock_method:
                    mock_method.return_value = File(open(TEST_IMAGE_PATH, "rb"))
                    # Not passing dimensions should result in file being read for it
                    tf = AsyncTempFile.create_for_field(
                        async_model_to_test._meta.get_field("image_with_dims"), "example.png"
                    )
                    tm = async_model_to_test.objects.create(
                        image_with_dims=AsyncFileInputData(key=tf.key, name="example.png")
                    )
                    self.assertEqual(tm.image_width, 16)
                    self.assertEqual(tm.image_height, 16)

                    mock_method.assert_called_once_with(tf.key, "rb")

    def test_image_field_dimensions_recalculate(self):
        # only test for parent model - we know that child model doesn't work for dimension recalculation.
        # this is because Django runs size retrieval at initialisation using a post_init signal. signals
        # don't activate for child models, and since the post_init signal is connected in the field's
        # internal contribute_to_class function, the signal is only attached to the parent model
        with mock.patch("test_alliance_platform_storage.storage.DummyStorage._open") as mock_method:
            tf = AsyncTempFile.create_for_field(
                AsyncFileTestModel._meta.get_field("image_with_dims"), "example.png"
            )
            tm = AsyncFileTestModel.objects.create(
                image_with_dims=AsyncFileInputData(key=tf.key, name=tf.original_filename, width=32, height=32)
            )
            self.assertEqual(tm.image_with_dims, "example.png")
            self.assertEqual(tm.image_width, 32)
            self.assertEqual(tm.image_height, 32)

            # Retrieving from database shouldn't trigger a calculation of dimensions either
            tm = AsyncFileTestModel.objects.get(pk=tm.pk)

            # If width & height are set using AsyncFileInputData then storage should not have to open the file
            mock_method.assert_not_called()

            mock_method.return_value = File(open(TEST_IMAGE_PATH, "rb"))
            tm.image_with_dims = "test.png"
            self.assertEqual(tm.image_width, 16)
            self.assertEqual(tm.image_height, 16)

            # refresh_from_db will trigger login in AsyncImageDescriptor
            mock_method.return_value = File(open(TEST_IMAGE_64_64_PATH, "rb"))
            # Assign same name but different underlying file; should recalculate width
            tm.image_with_dims = "test.png"

            self.assertEqual(tm.image_with_dims, "test.png")
            self.assertEqual(tm.image_width, 64)
            self.assertEqual(tm.image_height, 64)

            # Assigning after creation should trigger dimensions lookup
            mock_method.return_value = File(open(TEST_IMAGE_64_64_PATH, "rb"))
            tm = AsyncFileTestModel()
            tm.image_with_dims = "test.png"
            self.assertEqual(tm.image_with_dims, "test.png")
            self.assertEqual(tm.image_width, 64)
            self.assertEqual(tm.image_height, 64)

            # Using __init__ should trigger dimensions lookup
            mock_method.return_value = File(open(TEST_IMAGE_64_64_PATH, "rb"))
            tm = AsyncFileTestModel(image_with_dims="test.png")
            self.assertEqual(tm.image_with_dims, "test.png")
            self.assertEqual(tm.image_width, 64)
            self.assertEqual(tm.image_height, 64)

            # Using AsyncFileInputData with NO dimensions should trigger lookup
            mock_method.return_value = File(open(TEST_IMAGE_64_64_PATH, "rb"))
            tf = AsyncTempFile.create_for_field(
                AsyncFileTestModel._meta.get_field("image_with_dims"), "test.png"
            )
            tm = AsyncFileTestModel(image_with_dims=AsyncFileInputData(key=tf.key, name=tf.original_filename))
            self.assertEqual(tm.image_width, 64)
            self.assertEqual(tm.image_height, 64)

            # Using AsyncFileInputData WITH dimensions should NOT trigger lookup
            mock_method.reset_mock()
            tf = AsyncTempFile.create_for_field(
                AsyncFileTestModel._meta.get_field("image_with_dims"), "test.png"
            )
            tm = AsyncFileTestModel(
                image_with_dims=AsyncFileInputData(key=tf.key, name=tf.original_filename, width=16, height=16)
            )
            self.assertEqual(tm.image_width, 16)
            self.assertEqual(tm.image_height, 16)
            mock_method.assert_not_called()

    def test_image_field_dimensions_refresh_from_db(self):
        for async_model_to_test in self.MODELS_TO_TEST:
            with self.subTest(async_model_to_test.__name__):
                with mock.patch("test_alliance_platform_storage.storage.DummyStorage._open") as mock_method:
                    tf = AsyncTempFile.create_for_field(
                        async_model_to_test._meta.get_field("image_with_dims"), "example.png"
                    )
                    tm = async_model_to_test.objects.create(
                        image_with_dims=AsyncFileInputData(
                            key=tf.key, name=tf.original_filename, width=32, height=32
                        )
                    )
                    self.assertEqual(tm.image_with_dims, "example.png")
                    self.assertEqual(tm.image_width, 32)
                    self.assertEqual(tm.image_height, 32)

                    # If width & height are set using AsyncFileInputData then storage should not have to open the file
                    mock_method.assert_not_called()

                    tm2 = async_model_to_test.objects.get(pk=tm.pk)

                    mock_method.return_value = File(open(TEST_IMAGE_PATH, "rb"))
                    tm2.image_with_dims = "test.png"
                    tm2.save()
                    mock_method.assert_called_once_with("test.png", "rb")
                    mock_method.reset_mock()

                    # refresh_from_db will trigger logic in AsyncImageDescriptor that results in
                    # update_dimensions_being called. Technically this isn't necessary; if the
                    # record was retrieved direct from database instead of refresh_from_db it
                    # would not do this. Unfortunately we don't know about refresh_from_db in
                    # the field and so it just looks like the value has changed.
                    mock_method.return_value = File(open(TEST_IMAGE_PATH, "rb"))
                    tm.refresh_from_db()
                    mock_method.assert_called_once_with("test.png", "rb")

                    self.assertEqual(tm.image_with_dims, "test.png")
                    self.assertEqual(tm.image_width, 16)
                    self.assertEqual(tm.image_height, 16)

    def test_image_field_no_dimensions(self):
        with mock.patch("test_alliance_platform_storage.storage.DummyStorage._open") as mock_method:
            for async_model_to_test in self.MODELS_TO_TEST:
                with self.subTest(async_model_to_test.__name__):
                    tf = AsyncTempFile.create_for_field(
                        async_model_to_test._meta.get_field("image_no_dims"), "test.png"
                    )
                    tm = async_model_to_test.objects.create(
                        image_no_dims=AsyncFileInputData(key=tf.key, name=tf.original_filename)
                    )
                    self.assertEqual(tm.image_no_dims.name, "test.png")
                    mock_method.assert_not_called()

    def test_field_url(self):
        for async_model_to_test in self.MODELS_TO_TEST:
            with self.subTest(async_model_to_test.__name__):
                file1 = AsyncTempFile.create_for_field(
                    async_model_to_test._meta.get_field("file1"), "test.png"
                )
                file2 = AsyncTempFile.create_for_field(
                    async_model_to_test._meta.get_field("image_no_dims"), "example.png"
                )
                tm = async_model_to_test.objects.create(
                    file1=file1.key,
                    image_no_dims=AsyncFileInputData(key=file2.key, name=file2.original_filename),
                )
                field = async_model_to_test._meta.get_field("file1")
                registry = field.async_field_registry
                self.assertEqual(
                    tm.file1.url,
                    reverse(registry.attached_download_view)
                    + f"?field_id={registry.generate_id(field)}&instance_id={tm.pk}&_={shake_256(tm.file1.name.encode()).hexdigest(8)}",
                )
                field = async_model_to_test._meta.get_field("image_no_dims")
                registry = field.async_field_registry
                self.assertEqual(
                    tm.image_no_dims.url,
                    reverse(field.async_field_registry.attached_download_view)
                    + f"?field_id={registry.generate_id(field)}&instance_id={tm.pk}&_={shake_256(tm.image_no_dims.name.encode()).hexdigest(8)}",
                )

    def test_move_file(self):
        for async_model_to_test in self.MODELS_TO_TEST:
            with self.subTest(async_model_to_test.__name__):
                tf = AsyncTempFile.create_for_field(async_model_to_test._meta.get_field("file1"), "test.pdf")
                file = AsyncFileInputData(key=tf.key, name=tf.original_filename)
                tm = async_model_to_test.objects.create(file1=file)
                self.assertEqual(tm.file1.name, "test.pdf")

                # No error; key is same
                tm.file1 = AsyncFileInputData(key="test.pdf", name="test.pdf")
                tm.save()

                # Error; can't change key using AsyncFileInputData
                with self.assertRaisesRegex(ValidationError, "Invalid upload"):
                    with self.assertLogs(logging.getLogger("alliance_platform.storage")):
                        tm.file1 = AsyncFileInputData(key="test2.pdf", name="test.pdf")
                        tm.save()

                # No problem setting a new temp file location however
                tf = AsyncTempFile.create_for_field(async_model_to_test._meta.get_field("file1"), "test2.pdf")
                tm.file1 = AsyncFileInputData(key=tf.key, name=tf.original_filename)
                tm.save()
                self.assertEqual(tm.file1.name, "test2.pdf")

    def test_move_image_file(self):
        for async_model_to_test in self.MODELS_TO_TEST:
            with self.subTest(async_model_to_test.__name__):
                with mock.patch("test_alliance_platform_storage.storage.DummyStorage._open") as mock_method:
                    # On creation the temp file will be moved to the final location. As part
                    # of this dimensions should _not_ be calculated as they were passed in.
                    tf = AsyncTempFile.create_for_field(
                        async_model_to_test._meta.get_field("image_with_dims"), "test.png"
                    )
                    tm = async_model_to_test.objects.create(
                        image_with_dims=AsyncFileInputData(
                            key=tf.key, name=tf.original_filename, width=32, height=32
                        )
                    )
                    self.assertEqual(tm.image_with_dims.name, "test.png")
                    self.assertEqual(tm.image_width, 32)
                    self.assertEqual(tm.image_height, 32)

                    # If width & height are set using AsyncFileInputData then storage should not have to open the file
                    mock_method.assert_not_called()

                    # If temp file is created without passing the width/height then we
                    # do expect the dimensions to be calculated, but only once
                    tf = AsyncTempFile.create_for_field(
                        async_model_to_test._meta.get_field("image_with_dims"), "test.png"
                    )
                    mock_method.return_value = File(open(TEST_IMAGE_PATH, "rb"))
                    tm = async_model_to_test.objects.create(
                        image_with_dims=AsyncFileInputData(key=tf.key, name=tf.original_filename)
                    )

                    self.assertEqual(tm.image_with_dims, "test.png")
                    self.assertEqual(tm.image_width, 16)
                    self.assertEqual(tm.image_height, 16)

                    mock_method.assert_called_once_with(tf.key, "rb")
                    mock_method.mock_reset()

    def test_move_multiple_files(self):
        for i, async_model_to_test in enumerate(self.MODELS_TO_TEST):
            test_count = i + 1
            if async_model_to_test == AsyncFileTestModel:
                # 2 for transaction (savepoint + release), 1 for initial create, 1 for save on instance after files moved, 1 to select temp files, 1 to update temp files
                expected_queries = 6
            else:
                # 2 additional queries - 1 for create in child table, one for update in child table
                expected_queries = 8
            with self.subTest(async_model_to_test.__name__):
                # On creation the temp file will be moved to the final location. As part
                # of this dimensions should _not_ be calculated as they were passed in.
                image_with_dims = AsyncTempFile.create_for_field(
                    async_model_to_test._meta.get_field("image_with_dims"), "test.png"
                )
                image_no_dims = AsyncTempFile.create_for_field(
                    async_model_to_test._meta.get_field("image_no_dims"), "test2.png"
                )
                file1 = AsyncTempFile.create_for_field(
                    async_model_to_test._meta.get_field("file1"), "test.pdf"
                )
                with self.assertNumQueries(expected_queries):
                    tm = async_model_to_test.objects.create(
                        file1=AsyncFileInputData(key=file1.key, name=file1.original_filename),
                        image_no_dims=AsyncFileInputData(
                            key=image_no_dims.key, name=image_no_dims.original_filename
                        ),
                        image_with_dims=AsyncFileInputData(
                            key=image_with_dims.key,
                            name=image_with_dims.original_filename,
                            width=32,
                            height=32,
                        ),
                    )
                self.assertEqual(tm.image_with_dims.name, "test.png")
                self.assertEqual(tm.image_width, 32)
                self.assertEqual(tm.image_height, 32)

                self.assertEqual(tm.image_no_dims.name, "test2.png")
                self.assertEqual(tm.file1.name, "test.pdf")

                self.assertEqual(AsyncTempFile.objects.filter(moved_to_location="").count(), 0)
                self.assertEqual(
                    AsyncTempFile.objects.filter(moved_to_location__startswith="test").count(), 3 * test_count
                )

    def test_move_multiple_files_errors(self):
        for i, async_model_to_test in enumerate(self.MODELS_TO_TEST):
            test_count = i + 1
            if async_model_to_test == AsyncFileTestModel:
                # 2 for transaction (savepoint & release savepoint)
                # 1 for initial create, 1 for save on instance after files moved, 1 to select temp files, 1 to delete temp files and
                # 1 to save the error on the file1 AsyncTempFile
                expected_queries = 7
            else:
                # 2 additional queries - 1 for create in child table, one for update in child table
                expected_queries = 9
            with self.subTest(async_model_to_test.__name__):
                # On creation the temp file will be moved to the final location. As part
                # of this dimensions should _not_ be calculated as they were passed in.
                image_with_dims = AsyncTempFile.create_for_field(
                    async_model_to_test._meta.get_field("image_with_dims"), "test.png"
                )
                image_no_dims = AsyncTempFile.create_for_field(
                    async_model_to_test._meta.get_field("image_no_dims"), "test2.png"
                )
                file1 = AsyncTempFile.create_for_field(
                    async_model_to_test._meta.get_field("file1"), "test.pdf"
                )

                def fail_on_file1(self, from_key, to_key):
                    if to_key == "test.pdf":
                        raise ValueError("Cannot move")

                with mock.patch(
                    "test_alliance_platform_storage.storage.DummyStorage.move_file", new=fail_on_file1
                ):
                    with self.assertNumQueries(expected_queries):
                        with self.assertLogs(logging.getLogger("alliance_platform.storage")):
                            tm = async_model_to_test.objects.create(
                                file1=AsyncFileInputData(key=file1.key, name=file1.original_filename),
                                image_no_dims=AsyncFileInputData(
                                    key=image_no_dims.key, name=image_no_dims.original_filename
                                ),
                                image_with_dims=AsyncFileInputData(
                                    key=image_with_dims.key,
                                    name=image_with_dims.original_filename,
                                    width=32,
                                    height=32,
                                ),
                            )
                self.assertEqual(tm.image_with_dims.name, "test.png")
                self.assertEqual(tm.image_width, 32)
                self.assertEqual(tm.image_height, 32)

                self.assertEqual(tm.image_no_dims.name, "test2.png")
                self.assertEqual(tm.file1.name, file1.key)

                file1.refresh_from_db()

                self.assertEqual(AsyncTempFile.objects.filter(error__isnull=False).count(), 1 * test_count)
                self.assertEqual(
                    AsyncTempFile.objects.filter(moved_to_location__startswith="test").count(), 2 * test_count
                )
                self.assertRegex(file1.error, "Cannot move")


class AsyncFileFormTestCase(TestCase):
    def test_model_form_image(self):
        class AsyncFileTestModelForm(ModelForm):
            class Meta:
                model = AsyncFileTestModel
                fields = ["image_with_dims"]

        tf = AsyncTempFile.create_for_field(AsyncFileTestModel._meta.get_field("image_with_dims"), "test.png")
        key = tf.key
        form = AsyncFileTestModelForm(
            data={
                "image_with_dims": json.dumps(
                    {
                        "key": key,
                        "name": "test.png",
                        "width": 224,
                        "height": 550,
                    }
                )
            }
        )
        form.save()
        self.assertEqual(AsyncFileTestModel.objects.count(), 1)
        tm = AsyncFileTestModel.objects.first()
        self.assertEqual(tm.image_with_dims, tf.original_filename)
        self.assertEqual(tm.image_width, 224)
        self.assertEqual(tm.image_height, 550)

    def test_reject_bad_input(self):
        class AsyncFileTestModelForm(ModelForm):
            class Meta:
                model = AsyncFileTestModel
                fields = ["image_no_dims"]

        # Can't change the key on a field
        form = AsyncFileTestModelForm(data={"image_no_dims": "whatever.png"})
        self.assertFalse(form.is_valid())
        self.assertRegex(form.errors["image_no_dims"][0], "Bad input for file field")

        with self.assertLogs(logging.getLogger("alliance_platform.storage")):
            # Key isn't prefixed as a temporary path so must be rejected
            form = AsyncFileTestModelForm(
                data={"image_no_dims": json.dumps({"key": "/storage/test.png", "name": "test.png"})}
            )
            self.assertFalse(form.is_valid())
            self.assertRegex(form.errors["image_no_dims"][0], "Invalid upload received")

    def test_file_input(self):
        tm = AsyncFileTestModel.objects.create(image_no_dims="/storage/test.png")
        input = AsyncFileInput()
        field = tm._meta.get_field("image_no_dims")
        self.assertEqual(
            input.format_value(tm.image_no_dims),
            json.dumps(
                {
                    "key": "/storage/test.png",
                    "name": "test.png",
                    "url": f"/download-file/?field_id={field.async_field_registry.generate_id(field)}&instance_id={tm.pk}&_={shake_256(tm.image_no_dims.name.encode()).hexdigest(8)}",
                }
            ),
        )

    def test_file_input_max_len(self):
        class AsyncFileTestModelForm(ModelForm):
            class Meta:
                model = AsyncFileTestModel
                fields = ["image_no_dims"]

        name = "a" * (async_file_max_length + 1)
        key = f"{DummyStorage.temporary_key_prefix}/{name}"
        form = AsyncFileTestModelForm({"image_no_dims": json.dumps({"key": key, "name": name})})
        self.assertFalse(form.is_valid())
        self.assertRegex(
            form.errors["image_no_dims"][0],
            f"Ensure this value has at most {async_file_max_length} characters",
        )
        name = "a" * async_file_max_length
        key = f"{DummyStorage.temporary_key_prefix}/{name}"
        form = AsyncFileTestModelForm({"image_no_dims": json.dumps({"key": key, "name": name})})
        self.assertFalse(form.errors)

    def test_generate_temporary_path(self):
        storage = DummyStorage()
        temp_prefix_length = len("async-temp-files/2021/03/03/fVy5cSVBQpOb-")
        # Just enough to fit the extension
        length = temp_prefix_length + 4
        p = storage.generate_temporary_path("test.png", max_length=length)
        self.assertEqual(len(p), length)
        self.assertTrue(p.endswith("-.png"))

        # Can't fit the extension, error
        with self.assertRaises(SuspiciousFileOperation):
            storage.generate_temporary_path("test.png", max_length=length - 1)

        self.assertTrue(storage.generate_temporary_path("test.png").endswith("-test.png"))


class GenerateUploadUrlViewTestCase(TestCase):
    def _get_url(self, instance: AsyncFileTestModel | None = None, filename="abc.jpg"):
        field_id = default_async_field_registry.generate_id(AsyncFileTestModel._meta.get_field("file1"))
        query_params = dict(field_id=field_id, filename=filename)
        if instance:
            query_params["instanceId"] = instance.pk
        return f"{reverse(default_async_field_registry.attached_view)}?{urlencode(query_params)}"

    def _get_perm_url(
        self,
        field_name: Literal["file_no_perms", "file_custom_perms", "file_default_perms"],
        instance: AsyncFilePermTestModel | None = None,
        filename="abc.jpg",
    ):
        """Return a URL for the AsyncFilePermTestModel"""
        field_id = default_async_field_registry.generate_id(
            cast(AsyncFileField, AsyncFilePermTestModel._meta.get_field(field_name))
        )
        query_params = dict(fieldId=field_id, filename=filename)
        if instance:
            query_params["instanceId"] = instance.pk
        return f"{reverse(default_async_field_registry.attached_view)}?{urlencode(query_params)}"

    def test_validation(self):
        tests = [
            ({}, {"filename": ["This field is required."], "fieldId": ["This field is required."]}),
            ({"filename": "abc.jpg", "field_id": "invalid"}, {"fieldId": ["Unknown fieldId"]}),
        ]
        for params, errors in tests:
            with self.subTest(params=params):
                url = f"{reverse(default_async_field_registry.attached_view)}?{urlencode(params)}"
                response = self.client.get(url)
                self.assertEqual(response.status_code, 400)
                self.assertEqual(response.json(), errors)

    def test_no_permission_required(self):
        response = self.client.get(self._get_perm_url("file_no_perms"))
        self.assertEqual(response.status_code, 200)

    def test_perm_denied(self):
        with patch("test_alliance_platform_storage.models.User.has_perm", return_value=False):
            user = User.objects.create(username="test")
            self.client.force_login(user)
            response = self.client.get(self._get_perm_url("file_custom_perms"))
            self.assertEqual(response.status_code, 403)

    def test_custom_perm_create(self):
        with patch("test_alliance_platform_storage.models.User.has_perm", return_value=True) as mock_has_perm:
            user = User.objects.create(username="test")
            self.client.force_login(user)
            response = self.client.get(self._get_perm_url("file_custom_perms"))
            self.assertEqual(response.status_code, 200)
            mock_has_perm.assert_called_with("custom_create", None)

    def test_custom_perm_update(self):
        with patch("test_alliance_platform_storage.models.User.has_perm", return_value=True) as mock_has_perm:
            user = User.objects.create(username="test")
            self.client.force_login(user)
            existing_record = AsyncFilePermTestModel.objects.create()
            response = self.client.get(self._get_perm_url("file_custom_perms", instance=existing_record))
            self.assertEqual(response.status_code, 200)
            mock_has_perm.assert_called_with("custom_update", existing_record)

    def test_default_perm_create(self):
        with patch("test_alliance_platform_storage.models.User.has_perm", return_value=True) as mock_has_perm:
            user = User.objects.create(username="test")
            self.client.force_login(user)
            response = self.client.get(self._get_perm_url("file_default_perms"))
            self.assertEqual(response.status_code, 200)
            mock_has_perm.assert_called_with(
                "test_alliance_platform_storage.asyncfilepermtestmodel_create", None
            )

    def test_default_perm_update(self):
        with patch("test_alliance_platform_storage.models.User.has_perm", return_value=True) as mock_has_perm:
            user = User.objects.create(username="test")
            self.client.force_login(user)
            existing_record = AsyncFilePermTestModel.objects.create()
            response = self.client.get(self._get_perm_url("file_default_perms", instance=existing_record))
            self.assertEqual(response.status_code, 200)
            mock_has_perm.assert_called_with(
                "test_alliance_platform_storage.asyncfilepermtestmodel_update", existing_record
            )

    def test_generate_url_create(self):
        with patch("test_alliance_platform_storage.models.User.has_perm", return_value=True):
            user = User.objects.create(username="test")
            self.client.force_login(user)
            response = self.client.get(self._get_url())
            self.assertEqual(response.status_code, 200)
            data = json.loads(response.content)
            self.assertEqual(AsyncTempFile.objects.count(), 1)
            async_temp_file = AsyncTempFile.objects.first()
            assert async_temp_file is not None
            self.assertEqual(async_temp_file.original_filename, "abc.jpg")
            self.assertEqual(data["key"], async_temp_file.key)
            self.assertEqual(data["uploadUrl"], f"http://signme.com/{async_temp_file.key}")
            self.assertEqual(async_temp_file.field_name, "file1")
            self.assertEqual(async_temp_file.content_type.model_class(), AsyncFileTestModel)
            self.assertEqual(async_temp_file.moved_to_location, "")

            # Create record using temp file key. This results in move_file being
            # called, AsyncTempFile having `moved_to_location` set and the file field being updated
            # to the original filename
            file = AsyncFileTestModel.objects.create(file1=async_temp_file.key)
            self.assertEqual(file.file1, "abc.jpg")
            self.assertEqual(AsyncTempFile.objects.count(), 1)
            self.assertEqual(AsyncTempFile.objects.filter(moved_to_location="abc.jpg").count(), 1)

    def test_generate_url_update(self):
        with patch("test_alliance_platform_storage.models.User.has_perm", return_value=True):
            user = User.objects.create(username="test")
            self.client.force_login(user)
            instance = AsyncFileTestModel.objects.create(file1="some/value.jpg")
            response = self.client.get(self._get_url(instance=instance))
            self.assertEqual(response.status_code, 200)
            data = json.loads(response.content)
            self.assertEqual(AsyncTempFile.objects.count(), 1)
            async_temp_file = AsyncTempFile.objects.first()
            assert async_temp_file is not None
            self.assertEqual(async_temp_file.original_filename, "abc.jpg")
            self.assertEqual(data["key"], async_temp_file.key)
            self.assertEqual(data["uploadUrl"], f"http://signme.com/{async_temp_file.key}")
            self.assertEqual(async_temp_file.field_name, "file1")
            self.assertEqual(async_temp_file.content_type.model_class(), AsyncFileTestModel)
            self.assertEqual(async_temp_file.moved_to_location, "")

            # Update the file1 to use the new temp key. This results in move_file being
            # called, AsyncTempFile having `moved_to_location` set and the file field being updated
            # to the original filename
            instance.file1 = async_temp_file.key
            instance.save()
            self.assertEqual(AsyncTempFile.objects.count(), 1)
            self.assertEqual(AsyncTempFile.objects.filter(moved_to_location="abc.jpg").count(), 1)


class DownloadRedirectViewTestCase(TestCase):
    def _get_url(self, instance: AsyncFileTestModel, filename="abc.jpg"):
        field_id = default_async_field_registry.generate_id(AsyncFileTestModel._meta.get_field("file1"))
        query_params = dict(fieldId=field_id, filename=filename, instanceId=instance.pk)
        return f"{reverse(default_async_field_registry.attached_download_view)}?{urlencode(query_params)}"

    def _get_perm_url(
        self,
        instance: AsyncFilePermTestModel,
        field_name: Literal["file_no_perms", "file_custom_perms", "file_default_perms"],
        filename="abc.jpg",
    ):
        """Return a URL for the AsyncFilePermTestModel"""
        field_id = default_async_field_registry.generate_id(
            cast(AsyncFileField, AsyncFilePermTestModel._meta.get_field(field_name))
        )
        query_params = dict(fieldId=field_id, filename=filename, instanceId=instance.pk)
        return f"{reverse(default_async_field_registry.attached_download_view)}?{urlencode(query_params)}"

    def test_validation(self):
        tests = [
            ({}, {"filename": ["This field is required."], "fieldId": ["This field is required."]}),
            ({"filename": "abc.jpg", "field_id": "invalid"}, {"fieldId": ["Unknown fieldId"]}),
        ]
        for params, errors in tests:
            with self.subTest(params=params):
                url = f"{reverse(default_async_field_registry.attached_view)}?{urlencode(params)}"
                response = self.client.get(url)
                self.assertEqual(response.status_code, 400)
                self.assertEqual(response.json(), errors)

    def test_no_permission_required(self):
        record = AsyncFilePermTestModel.objects.create(file_no_perms="test.png")
        response = self.client.get(self._get_perm_url(record, "file_no_perms"))
        self.assertRedirects(response, "http://downloadme.com/test.png", fetch_redirect_response=False)

    def test_perm_denied(self):
        with patch("test_alliance_platform_storage.models.User.has_perm", return_value=False):
            user = User.objects.create(username="test")
            self.client.force_login(user)
            existing_record = AsyncFilePermTestModel.objects.create(file_no_perms="test.png")
            response = self.client.get(self._get_perm_url(existing_record, "file_custom_perms"))
            self.assertEqual(response.status_code, 403)

    def test_no_value(self):
        with patch("test_alliance_platform_storage.models.User.has_perm", return_value=True):
            user = User.objects.create(username="test")
            self.client.force_login(user)
            existing_record = AsyncFilePermTestModel.objects.create()
            response = self.client.get(self._get_perm_url(existing_record, "file_custom_perms"))
            self.assertEqual(response.status_code, 404)

    def test_custom_perm(self):
        with patch("test_alliance_platform_storage.models.User.has_perm", return_value=True) as mock_has_perm:
            user = User.objects.create(username="test")
            self.client.force_login(user)
            existing_record = AsyncFilePermTestModel.objects.create(file_custom_perms="test.png")
            response = self.client.get(self._get_perm_url(existing_record, "file_custom_perms"))
            mock_has_perm.assert_called_with("custom_detail", existing_record)
            self.assertRedirects(response, "http://downloadme.com/test.png", fetch_redirect_response=False)

    def test_default_perm(self):
        with patch("test_alliance_platform_storage.models.User.has_perm", return_value=True) as mock_has_perm:
            user = User.objects.create(username="test")
            self.client.force_login(user)
            existing_record = AsyncFilePermTestModel.objects.create(file_default_perms="test.png")
            response = self.client.get(self._get_perm_url(existing_record, "file_default_perms"))
            mock_has_perm.assert_called_with(
                "test_alliance_platform_storage.asyncfilepermtestmodel_detail", existing_record
            )
            self.assertRedirects(response, "http://downloadme.com/test.png", fetch_redirect_response=False)

    def test_redirects_to_download(self):
        with patch("test_alliance_platform_storage.models.User.has_perm", return_value=True):
            user = User.objects.create()
            record = AsyncFileTestModel.objects.create(file1="test.png")
            self.client.force_login(user=user)
            response = self.client.get(self._get_url(record, "test.png"))
            self.assertRedirects(response, "http://downloadme.com/test.png", fetch_redirect_response=False)



import json
from json import JSONDecodeError
from typing import Any
from typing import Protocol
from typing import cast

from alliance_platform.storage.async_uploads.models import AsyncTempFile
from alliance_platform.storage.async_uploads.registry import AsyncFieldRegistry
from alliance_platform.storage.async_uploads.registry import default_async_field_registry
from allianceutils.util import camelize
from allianceutils.util import underscoreize
from django.core.exceptions import BadRequest
from django.core.exceptions import PermissionDenied
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Manager
from django.http import HttpRequest
from django.http import HttpResponse
from django.http import HttpResponseNotFound
from django.http import HttpResponseRedirect
from django.http import JsonResponse
from django.http import QueryDict
from django.shortcuts import get_object_or_404
from django.views import View


def validate_generate_upload_url(data: QueryDict, registry: AsyncFieldRegistry):
    field_id = data.get("field_id")
    filename = data.get("filename")
    params = data.get("params", None)
    instance_id = data.get("instance_id", None)
    errors: dict[str, str] = {}

    if not field_id:
        errors["field_id"] = "This field is required."
    elif field_id not in registry.fields_by_id:
        errors["field_id"] = "Unknown fieldId"
    if not filename:
        errors["filename"] = "This field is required."
    if errors:
        raise ValidationError(errors)
    if params:
        try:
            params = json.loads(params)
        except JSONDecodeError:
            errors["params"] = "Invalid JSON"

    return {
        "field_id": field_id,
        "filename": filename,
        "params": params,
        "instance_id": instance_id,
    }


class ViewProtocol(Protocol):
    view_initkwargs: dict[str, Any]  # django-stubs is missing this


class ModelWithDefaultManager(Protocol):
    objects: Manager


class GenerateUploadUrlView(View):
    """View to generate a URL using :meth:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.generate_upload_url`.

    Each view is tied to a specific registry which you can specify in the ``registry`` kwarg (defaults to :data:`~alliance_platform.storage.async_uploads.registry.default_async_field_registry`).

    This view expects a ``async_field_id``, a ``filename`` and optionally an ``instance_id`` if it's an update for an existing
    record. ``async_field_id`` is used to look up in ``registry`` the matching ``field`` instance. From the retrieved field
    instance the corresponding :class:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage` is retrieved along with the permissions
    to check (see :class:`~alliance_platform.storage.async_uploads.models.AsyncFileMixin` for where to specify this).

    If the permission check passes a :class:`~alliance_platform.storage.async_uploads.models.AsyncTempFile` record is created which stores the
    ``filename`` and a generated key that will be used as a temporary location to upload the file to. An upload URL
    returned from :meth:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.generate_upload_url` and the temporary key is returned
    to the frontend. The frontend will submit the returned temporary key from the form and save it on the target model.
    On save the file is then moved to the permanent location using :meth:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.move_file`
    (this happens as part of :class:`~alliance_platform.storage.async_uploads.models.AsyncFileMixin`)

    See :class:`~alliance_platform.storage.async_uploads.models.AsyncFileMixin` for a detailed explanation of how all the pieces fit together.

    This view is automatically registered if you have followed the :ref:`register-urls` guide.
    """

    registry = default_async_field_registry

    @classmethod
    def as_view(cls, **initkwargs: Any):
        view: ViewProtocol = cast(ViewProtocol, super().as_view(**initkwargs))
        registry = view.view_initkwargs.get("registry", default_async_field_registry)
        if registry.attached_view:
            import warnings

            warnings.warn(
                f"{registry} has already been attached to one GenerateUploadUrlView. Pass a different registry in like `GenerateUploadUrlView.as_view({{ registry: another_reg }})`"
            )
        registry.attached_view = view
        return view

    def get(self, request: HttpRequest) -> HttpResponse:
        try:
            validated_data = validate_generate_upload_url(underscoreize(request.GET), self.registry)
        except ValidationError as e:
            return JsonResponse(camelize(e.message_dict), status=400)
        field_id = validated_data["field_id"]
        field = self.registry.fields_by_id[field_id]
        instance_id = validated_data["instance_id"]
        obj = None
        if instance_id:
            manager = cast(ModelWithDefaultManager, field.model).objects
            obj = manager.get(pk=instance_id)
        required_perm = field.perm_update if obj else field.perm_create
        if required_perm is not None and not request.user.has_perm(required_perm, obj):
            raise PermissionDenied

        filename = validated_data["filename"]
        if field.file_restrictions:
            if not field.is_valid_filetype(filename):
                raise BadRequest

        temp_file = AsyncTempFile.create_for_field(field, filename)
        conditions = []
        if field.max_size:
            conditions.append(["content-length-range", 0, field.max_size * 1024 * 1024])
        # this matches any, but is OK because we are signing a single key only anyway which will restrict the type
        conditions.append(["starts-with", "$Content-Type", ""])
        return JsonResponse(
            {
                "uploadUrl": field.storage.generate_upload_url(
                    temp_file.key,
                    field_id,
                    fields=validated_data.get("params"),
                    conditions=conditions or None,
                ),
                "key": temp_file.key,
            }
        )


def validate_async_file_download(data: QueryDict, registry: AsyncFieldRegistry):
    field_id = data.get("field_id")
    instance_id = data.get("instance_id", None)

    if not field_id:
        raise ValidationError({"field_id": "This field is required."})

    if field_id not in registry.fields_by_id:
        raise ValidationError({"field_id": "Unknown fieldId"})

    return {"field_id": field_id, "instance_id": instance_id}


class DownloadRedirectView(View):
    """View that checks permissions and redirects to a download URL using :meth:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.generate_download_url`.

    Each view is tied to a specific registry which you can specify in the ``registry`` kwarg (defaults to :data:`~alliance_platform.storage.async_uploads.registry.default_async_field_registry`).

    This view expects a ``async_field_id`` and ``instance_id`` which is the primary key of the record the file field is
    attached to. record. ``async_field_id`` is used to look up in ``registry`` the matching ``field`` instance. From the
    retrieved field instance the corresponding :class:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage` is retrieved along
    with the permission (``perm_download``) to check (see :class:`~alliance_platform.storage.async_uploads.models.AsyncFileMixin` for
    where to specify this).

    Once the permission check has passed the view will redirect to the URL returned from
    :meth:`~alliance_platform.storage.async_uploads.storage.base.AsyncUploadStorage.generate_download_url`.

    See :class:`~alliance_platform.storage.async_uploads.models.AsyncFileMixin` for a detailed explanation of how all the pieces fit together.

    This view is automatically registered if you have followed the :ref:`register-urls` guide.
    """

    registry = default_async_field_registry

    @classmethod
    def as_view(cls, **initkwargs: Any):
        view: ViewProtocol = cast(ViewProtocol, super().as_view(**initkwargs))
        registry = view.view_initkwargs.get("registry", default_async_field_registry)
        if registry.attached_download_view:
            import warnings

            warnings.warn(
                f"{registry} has already been attached to one AsyncFileDownloadView. Pass a different registry in like `AsyncFileDownloadView.as_view({{ registry: another_reg }})`"
            )
        registry.attached_download_view = view
        return view

    def get(self, request: HttpRequest):
        try:
            validated_data = validate_async_file_download(underscoreize(request.GET), self.registry)
        except ValidationError as e:
            return JsonResponse(camelize(e.message_dict), status=400)
        field_id = validated_data["field_id"]
        field = self.registry.fields_by_id[field_id]
        instance_id = validated_data.get("instance_id")

        manager = cast(ModelWithDefaultManager, field.model).objects
        obj: models.Model = get_object_or_404(manager, pk=instance_id)
        if field.perm_detail is not None and not request.user.has_perm(field.perm_detail, obj):
            raise PermissionDenied
        value = getattr(obj, field.name)
        if not value:
            return HttpResponseNotFound(f"No value set for {field.name}")
        return HttpResponseRedirect(field.storage.generate_download_url(value.name, field_id))



import os
import shutil
import tempfile

from django.conf import STATICFILES_STORAGE_ALIAS
from django.conf import settings
from django.contrib.staticfiles import storage
from django.core.management import call_command
from django.template import Context
from django.template import Template
from django.test import SimpleTestCase
from django.test import override_settings


@override_settings(
    STORAGES={
        **settings.STORAGES,
        STATICFILES_STORAGE_ALIAS: {
            "BACKEND": "alliance_platform.storage.staticfiles.storage.ExcludingManifestStaticFilesStorage",
            "OPTIONS": {"exclude_patterns": ["manifest_test/build/*", "*.pdf"]},
        },
    },
)
class TestExcludingManifestStaticFilesStorage(SimpleTestCase):
    """
    Tests for ExcludingManifestStaticFilesStorage
    """

    def setUp(self):
        super().setUp()
        temp_dir = tempfile.mkdtemp()
        # Override the STATIC_ROOT for all tests from setUp to tearDown
        # rather than as a context manager
        self.static_root = temp_dir
        patched_settings = self.settings(STATIC_ROOT=temp_dir)
        patched_settings.enable()
        call_command(
            "collectstatic",
            interactive=False,
            verbosity=0,
        )
        self.addCleanup(shutil.rmtree, temp_dir)
        self.addCleanup(patched_settings.disable)

    def render_template(self, template, **kwargs):
        if isinstance(template, str):
            template = Template(template)
        return template.render(Context(**kwargs)).strip()

    def static_template_snippet(self, path, asvar=False):
        if asvar:
            return "{%% load static from static %%}{%% static '%s' as var %%}{{ var }}" % path
        return "{%% load static from static %%}{%% static '%s' %%}" % path

    def hashed_file_path(self, path):
        fullpath = self.render_template(self.static_template_snippet(path))
        return fullpath.removeprefix(settings.STATIC_URL)

    def test_hash_ignored(self):
        self.assertEqual(
            self.hashed_file_path("manifest_test/build/main-efg456.js"), "manifest_test/build/main-efg456.js"
        )
        self.assertEqual(
            self.hashed_file_path("manifest_test/build/body-abc123.png"),
            "manifest_test/build/body-abc123.png",
        )
        self.assertEqual(self.hashed_file_path("manifest_test/info.pdf"), "manifest_test/info.pdf")

    def test_excluded_files_not_hashed(self):
        # build/main-efg456.js should not be hashed
        relpath = self.hashed_file_path("manifest_test/build/main-efg456.js")
        self.assertEqual(relpath, "manifest_test/build/main-efg456.js")

        # build/body-abc123.png should not be hashed
        relpath = self.hashed_file_path("manifest_test/build/body-abc123.png")
        self.assertEqual(relpath, "manifest_test/build/body-abc123.png")

        # info.pdf should not be hashed
        relpath = self.hashed_file_path("manifest_test/info.pdf")
        self.assertEqual(relpath, "manifest_test/info.pdf")

    def test_manifest(self):
        # Check that the manifest file exists and contains the correct mappings
        manifest_path = os.path.join(self.static_root, "staticfiles.json")
        with open(manifest_path, "r") as f:
            manifest = f.read()
            # Check that hashed files are in the manifest
            self.assertIn('"manifest_test/css/base.css"', manifest)
            self.assertIn('"manifest_test/css/base.', manifest)
            self.assertIn('"manifest_test/images/header.png"', manifest)
            self.assertIn('"manifest_test/images/header.', manifest)
            # Check that excluded files are not hashed in the manifest
            self.assertNotIn('"manifest_test/build/main-efg456.js"', manifest)
            self.assertNotIn('"manifest_test/build/main-efg456.', manifest)
            self.assertNotIn('"manifest_test/info.pdf"', manifest)
            self.assertNotIn('"manifest_test/info.', manifest)

    def test_non_hashed_references(self):
        relpath = self.hashed_file_path("manifest_test/css/base.css")
        self.assertRegex(relpath, r"^manifest_test/css/base\.[0-9a-f]{12}\.css$")
        # Read the content of the collected css file
        with storage.staticfiles_storage.open(relpath) as f:
            content = f.read()
            # Check that reference to header.png is hashed
            self.assertRegex(content.decode(), r"\.\./images/header\.[0-9a-f]{12}\.png")
            # Check that reference to body-abc123.png is not hashed (excluded)
            self.assertIn("../build/body-abc123.png", content.decode())

    def test_collectstatic_output(self):
        # Run collectstatic again and capture the output
        from io import StringIO

        shutil.rmtree(self.static_root)
        out = StringIO()
        call_command("collectstatic", interactive=False, verbosity=2, stdout=out)
        output = out.getvalue()
        # Check that excluded files are not post-processed
        self.assertIn("Post-processed 'manifest_test/css/base.css'", output)
        self.assertIn("Post-processed 'manifest_test/images/header.png'", output)
        self.assertNotIn("Post-processed 'manifest_test/build/main-efg456.js'", output)
        self.assertNotIn("Post-processed 'manifest_test/info.pdf'", output)



from datetime import timedelta
from typing import Any

from alliance_platform.storage.async_uploads.storage.base import AsyncUploadStorage
from alliance_platform.storage.async_uploads.storage.base import GenerateUploadUrlResponse
from alliance_platform.storage.settings import ap_storage_settings
from django.core.exceptions import ImproperlyConfigured
from django.utils import timezone

try:
    from storages.backends.azure_storage import (  # type: ignore[import-untyped] # no types for storages
        AzureStorage,
    )
except ImportError as e:
    raise ImproperlyConfigured(
        "Optional dependency 'django-storages[azure]' is not installed. This is required for the AzureAsyncUploadStorage backend."
    ) from e

try:
    from azure.storage.blob import BlobClient
    from azure.storage.blob import BlobSasPermissions
    from azure.storage.blob import generate_blob_sas
except ImportError as e:
    raise ImproperlyConfigured(
        "Optional dependency 'azure-storage-blob' is not installed. This is required for the AzureAsyncUploadStorage backend."
    ) from e


class AzureAsyncUploadStorage(AzureStorage, AsyncUploadStorage):
    """Azure implementation of AsyncUploadStorage

    Uses signed URLs for uploading.
    """

    def generate_upload_url(
        self,
        name: str,
        field_id: str,
        *,
        expire: int = ap_storage_settings.UPLOAD_URL_EXPIRY,
        conditions: Any | None = None,
        fields: Any | None = None,
    ) -> GenerateUploadUrlResponse:
        """
        Generates a presigned PUT signed URL. Returns a dictionary with two elements: url and fields. Url is the url to post to. Fields is a dictionary filled with the form fields and respective values to use when submitting the post.

        ``conditions`` and ``fields`` currently ignored - here for compat with existing code that was written
        with S3 in mind

        Args:
            name:       the key for the upload
            expire:     time until presigned POST expires
            conditions: Ignored currently
            fields:     Ignored currently
        """
        # fields = fields.copy() if fields else {}
        # conditions = conditions.copy() if conditions else {}
        credential = generate_blob_sas(
            self.account_name,
            self.azure_container,
            name,
            account_key=self.account_key,
            permission=BlobSasPermissions(create=True),
            expiry=timezone.now() + timedelta(seconds=expire),
        )
        container_blob_url = self.client.get_blob_client(name).url
        return {"url": BlobClient.from_blob_url(container_blob_url, credential=credential).url, "fields": {}}

    def generate_download_url(
        self, key: str, field_id: str, expire=ap_storage_settings.DOWNLOAD_URL_EXPIRY, **kwargs
    ):
        """Generates a signed URL to download the file"""
        return super().url(key, expire=expire)  # type: ignore[call-arg] # Not sure why this is an error... AzureStorage definitely supports expire

    def move_file(self, from_key, to_key):
        """Moves file by copying :code:`from_key` to :code:`to_key` and then deletes :code:`from_key`"""
        blob_client = self.client.get_blob_client(from_key)
        blob_url = blob_client.url
        to_blob_client = self.client.get_blob_client(to_key)
        response = to_blob_client.start_copy_from_url(blob_url)
        if response["copy_status"] == "success":
            blob_client.delete_blob()



from typing import Any

from alliance_platform.storage.async_uploads.storage.base import AsyncUploadStorage
from alliance_platform.storage.async_uploads.storage.base import GenerateUploadUrlResponse
from alliance_platform.storage.settings import ap_storage_settings
from django.core.exceptions import ImproperlyConfigured

try:
    from storages.backends.s3 import S3Storage  # type: ignore[import-untyped] # no types for storages
except ImportError as e:
    raise ImproperlyConfigured(
        "Optional dependency 'django-storages[s3]' is not installed. This is required for the S3AsyncUploadStorage backend."
    ) from e


class S3AsyncUploadStorage(S3Storage, AsyncUploadStorage):
    """S3 implementation of AsyncUploadStorage

    Uses signed URLs for uploading.
    """

    # if the bucket is public and you don't want a querystring attached, set this to False.
    querystring_auth = True

    def generate_upload_url(
        self,
        name: str,
        field_id: str,
        *,
        expire: int | None = ap_storage_settings.UPLOAD_URL_EXPIRY,
        conditions: Any | None = None,
        fields: Any | None = None,
    ) -> GenerateUploadUrlResponse:
        """
        Generates a presigned POST signed URL. Returns a dictionary with two elements: url and fields. Url is the url to post to. Fields is a dictionary filled with the form fields and respective values to use when submitting the post.
        e.g

        .. code-block:: json

            {
                "url": "https://mybucket.s3.amazonaws.com",
                "fields": {
                    "acl": "public-read",
                    "key": "mykey",
                    "signature": "mysignature",
                    "policy": "mybase64 encoded policy"
                }
            }

        Args:
            name:       the key for the upload
            expire:     time until presigned POST expires
            conditions: A list of conditions to include in the policy. See https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.generate_presigned_post for valid values.
                        Conditions that are included may pertain to acl, content-length-range, Cache-Control, Content-Type, Content-Disposition, Content-Encoding, Expires, success_action_redirect, redirect, success_action_status, and/or x-amz-meta-.
            fields:     A dictionary of prefilled form fields to build on top of.
                        Elements that may be included are acl, Cache-Control, Content-Type, Content-Disposition, Content-Encoding, Expires, success_action_redirect, redirect, success_action_status, and x-amz-meta-.
                        See https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.generate_presigned_post for valid values.
        """
        fields = fields.copy() if fields else {}
        conditions = conditions.copy() if conditions else {}
        return self.bucket.meta.client.generate_presigned_post(
            self.bucket_name, name, Fields=fields, Conditions=conditions, ExpiresIn=expire
        )

    def generate_download_url(
        self, key: str, field_id: str, expire=ap_storage_settings.DOWNLOAD_URL_EXPIRY, **kwargs
    ):
        """Generates a signed URL to download the file"""
        return super().url(key, expire=expire)  # type: ignore[call-arg] # Not sure why this is an error... S3Storage definitely supports expire

    def move_file(self, from_key, to_key):
        """Moves file by copying :code:`from_key` to :code:`to_key` and then deletes :code:`from_key`"""
        object = {"Bucket": self.bucket_name, "Key": from_key}
        client = self.bucket.meta.client
        client.copy(object, self.bucket_name, to_key)
        client.delete_object(**object)



from django.db import migrations


def copy_data_from_common_storage_table(apps, schema_editor):
    # Check if the old table exists
    sql = """
        DO $$ 
        BEGIN
            -- Check if the old table exists
            IF EXISTS (
                SELECT FROM information_schema.tables 
                WHERE table_name = 'common_storage_async_temp_file' 
                AND table_schema = 'public'
            ) THEN
                -- Copy data from the old table to the new one
                INSERT INTO alliance_platform_storage_async_temp_file (
                    created_at, original_filename, key, field_name, content_type_id, error, moved_to_location
                )
                SELECT 
                    created_at, original_filename, key, field_name, content_type_id, error, moved_to_location
                FROM common_storage_async_temp_file;
            END IF;
        END $$;
        """

    # Execute the SQL
    schema_editor.execute(sql)


class Migration(migrations.Migration):
    dependencies = [
        ("alliance_platform_storage", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(copy_data_from_common_storage_table),
    ]
